# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
System-agnostic convenient functions.
"""
import warnings
from typing import (
    Dict,
    List,
    Tuple,
    Union,
)

import numpy as np
from qctrlcommons.preconditions import (
    check_argument,
    check_argument_iterable,
    check_argument_positive_integer,
)
from scipy.linalg import fractional_matrix_power

# https://github.com/PyCQA/pylint/issues/3744
from scipy.special import betaincinv  # pylint: disable=no-name-in-module

from qctrltoolkit.namespace import Namespace
from qctrltoolkit.toolkit_utils import expose


@expose(Namespace.UTILS)
def confidence_ellipse_matrix(
    hessian: np.ndarray,
    cost: float,
    measurement_count: int,
    confidence_fraction: float = 0.95,
):
    r"""
    Calculates a matrix that you can use to represent the confidence region
    of parameters that you estimated. Pass to this function the Hessian of
    the residual sum of squares with respect to the parameters, and use the
    output matrix to transform a hypersphere into a hyperellipse representing
    the confidence region. You can then plot this hyperellipse to visualize the
    confidence region.

    Alternatively, you can apply a (2,2)-slice of the transformation matrix to
    a unit circle to visualize the confidence ellipse for a pair of estimated
    parameters.

    Parameters
    ----------
    hessian : np.ndarray
        The Hessian of the residual sum of squares cost with respect to the estimated parameters,
        :math:`H`.
        Must be a square matrix.
    cost : float
        The residual sum of squares of the measurements with respect to the actual measurements,
        :math:`C_\mathrm{RSS}`.
        Must be positive.
    measurement_count : int
        The number of measured data points, :math:`n`.
        Must be positive.
    confidence_fraction : float, optional
        The confidence fraction for the ellipse, :math:`\alpha`.
        If provided, must be between 0 and 1.
        Defaults to 0.95.

    Returns
    -------
    np.ndarray
        A :math:`(p, p)`-matrix which transforms a unit hypersphere in a p-dimensional
        space into a hyperellipse representing the confidence region for the
        confidence fraction :math:`\alpha`. Here :math:`p` is the dimension of the Hessian matrix.

    Notes
    -----
    From the Hessian matrix of the residual sum of squares with respect
    to the estimated parameters :math:`\{\lambda_i\}`,

    .. math::
        H_{ij} = \frac{\partial^2 C_\mathrm{RSS}}{\partial \lambda_i \partial \lambda_j} \; ,

    we can estimate the covariance matrix for the estimated parameters

    .. math::
        \Sigma = \left( \frac{n-p}{2 C_\mathrm{RSS}} H \right)^{-1}  \; .

    Finally, we can find a scaling factor :math:`z`, such that the matrix :math:`z \Sigma^{1/2}`
    transforms the coordinates of a unit hypersphere in a p-dimensional space into a
    hyperellipse representing the confidence region

    .. math::
        z = \sqrt{p F_{1-\alpha} \left( \frac{n-p}{2}, \frac{p}{2} \right)} \; ,

    where :math:`F_{1-\alpha}(a,b)` is the point of the `F-distribution
    <https://en.wikipedia.org/wiki/F-distribution>`_ where the probability in
    the tail is equal to :math:`F_{1-\alpha}(a,b)`.

    For more details, see the topic
    `Characterizing your hardware using system identification in Boulder Opal
    <https://docs.q-ctrl.com/boulder-opal/topics/characterizing-your-hardware-using-system-identification-in-boulder-opal>`_
    and `N. R. Draper and I. Guttman, The Statistician 44, 399 (1995)
    <https://doi.org/10.2307/2348711>`_.
    """

    parameter_count = hessian.shape[0]

    check_argument(
        hessian.shape == (parameter_count, parameter_count),
        "Hessian must be a square matrix.",
        {"hessian": hessian},
    )

    check_argument(cost > 0, "The cost must be positive.", {"cost": cost})

    check_argument_positive_integer(measurement_count, "measurement_count")

    check_argument(
        0 < confidence_fraction < 1,
        "The confidence fraction must be between 0 and 1.",
        {"confidence_fraction": confidence_fraction},
    )

    # Estimate covariance matrix from the Hessian.
    covariance_matrix = np.linalg.inv(
        0.5 * hessian * (measurement_count - parameter_count) / cost
    )

    # Calculate scaling factor for the confidence region.
    iibeta = betaincinv(
        (measurement_count - parameter_count) / 2,
        parameter_count / 2,
        1 - confidence_fraction,
    )
    inverse_cdf = (
        (measurement_count - parameter_count) / parameter_count * (1 / iibeta - 1)
    )
    scaling_factor = np.sqrt(parameter_count * inverse_cdf)

    # Calculate confidence region for the confidence fraction.
    return scaling_factor * fractional_matrix_power(covariance_matrix, 0.5)


@expose(Namespace.UTILS)
def pwc_arrays_to_pairs(
    durations: Union[float, np.ndarray], values: np.ndarray
) -> List[Dict[str, np.ndarray]]:
    r"""
    Creates a list of dictionaries with "value" and "duration" keys representing
    a piecewise-constant function from arrays containing the durations and values.

    You can use this function to prepare a control to be plotted with the
    `plot_controls` function from the Q-CTRL Visualizer package.

    Parameters
    ----------
    durations : np.ndarray or float
        The durations of the PWC segments as a 1D array or as a float.
        If a single (float) value is passed, it is taken as the total duration of the PWC and
        all segments are assumed to have the same duration.
    values : np.ndarray
        The values of the PWC.

    Returns
    -------
    list[dict]
        A list of dictionaries (with "value" and "duration" keys) defining a PWC.

    Examples
    --------
    >>> qctrl.utils.pwc_arrays_to_pairs(1.0, np.array([3,-2,4,1]))
    [
        {'duration': 0.25, 'value': 3},
        {'duration': 0.25, 'value': -2},
        {'duration': 0.25, 'value': 4},
        {'duration': 0.25, 'value': 1}
    ]

    Plotting a control using the function plot_controls from the Q-CTRL Visualizer package.

    >>> plot_controls(
    ...     plt.figure(),
    ...     {"control": qctrl.utils.pwc_arrays_to_pairs(np.array([1,3,2]), np.array([-1,1,2]))},
    ... )
    """
    check_argument(
        np.all(durations > 0.0), "Durations must be positive.", {"durations": durations}
    )

    if np.ndim(durations) == 0:
        durations = np.repeat(durations / values.shape[0], values.shape[0])
    else:
        check_argument(
            np.ndim(durations) == 1,
            "Durations must be a float or 1D array.",
            {"durations": durations},
        )
        check_argument(
            values.shape[0] == len(durations),
            "Durations must be a float or a 1D array with the same length as values.",
            {"durations": durations, "values": values},
        )

    return [{"duration": d, "value": v} for d, v in zip(durations, values)]


@expose(Namespace.UTILS)
def pwc_pairs_to_arrays(pwc: List) -> Tuple[np.ndarray, np.ndarray, int]:
    r"""
    Extracts arrays with the durations and values representing a piecewise-constant function
    from a list of dictionaries with "value" and "duration" keys.

    You can use this function to retrieve the durations and values of a PWC extracted from
    a Boulder Opal graph calculation.

    Parameters
    ----------
    pwc : list[...list[list[dict]]]
        A nested list of lists of ... of list of dictionaries.
        The outer lists represent batches.
        The dictionaries in the innermost list must have "value" and "duration" keys, defining a
        PWC.

    Returns
    -------
    np.ndarray
        The durations of the PWC.
    np.ndarray
        The values of the PWC.
    int
        The number of batch dimensions.

    Examples
    --------
    >>> pwc_example = [
                 {'duration': 1.0, 'value': 3},
                 {'duration': 0.5, 'value': -2},
                 {'duration': 0.5, 'value': 4},
                 ]
    >>> qctrl.utils.pwc_pairs_to_arrays(pwc_example)
        (array([1., 0.5, 0.5]), array([3, -2, 4]), 0)

    >>> pwc_example = [
                [{'duration': 1.0, 'value': 3},
                 {'duration': 0.5, 'value': -2},
                 {'duration': 0.5, 'value': 4}],
                [{'duration': 1.0, 'value': 2},
                 {'duration': 0.5, 'value': 3},
                 {'duration': 0.5, 'value': -1}]
                 ]
    >>> qctrl.utils.pwc_pairs_to_arrays(pwc_example)
        (array([1., 0.5, 0.5]),
         array([[3, -2, 4],
                [2,  3, -1]]),
         1)

    Defining a PWC from a graph calculation.

    >>> graph.pwc(*qctrl.utils.pwc_pairs_to_arrays(result.output['signal']))
        <Pwc: name="pwc_#1", operation_name="pwc", value_shape=(), batch_shape=(4, 3)>
    """

    def extract_batch(pwc_list):
        check_argument(
            isinstance(pwc_list, list),
            "Item in outer lists is not a list.",
            {"pwc": pwc},
        )

        _type = type(pwc_list[0])

        check_argument(
            all(isinstance(e, _type) for e in pwc_list),
            "Items in same batch dimension not of same type.",
            {"pwc": pwc},
        )

        if isinstance(pwc_list[0], list):
            _len = len(pwc_list[0])
            check_argument(
                all(len(e) == _len for e in pwc_list),
                "Lists in same batch dimension not of same length.",
                {"pwc": pwc},
            )

            # The batching goes one level deeper.
            res = [extract_batch(k) for k in pwc_list]
            return [d for d, _ in res], [v for _, v in res]

        # We've reached the PWC list of dictionaries.
        for pwc_dict in pwc_list:
            check_argument(
                isinstance(pwc_dict, dict),
                "Item in innermost list not a dictionary.",
                {"pwc": pwc},
            )
            check_argument(
                "duration" in pwc_dict.keys() and "value" in pwc_dict.keys(),
                "`duration` or `value` key missing in given PWC.",
                {"pwc": pwc},
            )

        return [s["duration"] for s in pwc_list], [s["value"] for s in pwc_list]

    durations, values = extract_batch(pwc)

    # Durations and values have the same shape due to the input structure.
    # Check that the number of segments for all the PWCs are the same and all batches in
    # the same batch dimension must have the same number of elements.
    # If either of these conditions is not the case the data type (np.dtype) of
    # durations and values will be an object.
    # Note that this test is still necessary as the argument (where pwc is some dictionary with
    # duration and value keys)
    #    [[pwc], [[pwc, pwc]]]
    # will pass the checks in extract_batch.
    with warnings.catch_warnings(record=True) as warning_records:
        durations = np.array(durations)

    check_argument(
        len(warning_records) == 0,
        "The number of segments of each batch element must be the same, "
        "and all batches in the same batch dimension must have the same number of elements.",
        {"pwc": pwc},
    )
    values = np.array(values)
    time_dimension = durations.ndim - 1

    durations = np.reshape(durations, [-1, durations.shape[-1]])

    check_argument(
        np.all(np.diff(durations, axis=0) == 0),
        "The durations of all batch elements must be the same.",
        {"pwc": pwc},
        extras={"durations": durations},
    )

    durations = durations[0]

    check_argument(
        np.all(durations > 0.0),
        "Durations must be positive.",
        {"pwc": pwc},
        extras={"durations": durations},
    )

    return durations, values, time_dimension


@expose(Namespace.UTILS)
def extract_filter_function_arrays(samples: List) -> Tuple[np.ndarray]:
    """
    Convert the output samples of a filter function calculation into NumPy arrays.

    Each sample contains a frequency value and the corresponding inverse power,
    inverse power uncertainty, and frequency domain noise operator. The four returned
    NumPy arrays are the frequencies, inverse powers, inverse power uncertainties,
    and the frequency domain noise operators. If any of the values is None this function
    returns None rather than the corresponding NumPy array.

    Parameters
    ----------
    samples : list[qctrl.types.filter_function.Sample]
        The samples of a ``qctrl.functions.calculate_filter_function`` result.

    Returns
    -------
    np.ndarray or None
        The 1D NumPy array of frequencies.
        Returns None if any of the frequencies are None.
    np.ndarray or None
        The 1D NumPy array of inverse powers.
        Returns None if any of the inverse powers are None.
    np.ndarray or None
        The 1D NumPy array of inverse power uncertainties.
        Returns None if any of the inverse power uncertainties are None.
    np.ndarray or None
        The 3D NumPy array of frequency domain noise operators.
        Returns None if any of the frequency domain noise operators are None.

    Raises
    ------
    QctrlArgumentsValueError
        If samples is not an iterable or a sample is missing the frequency, inverse power,
        inverse power uncertainty, or frequency domain noise operator.

    Examples
    --------
    >>> samples = [
            Sample(
                frequency=0.06283185307179587,
                inverse_power=2.469046857168369,
                inverse_power_uncertainty=0.0016463055859327047,
                frequency_domain_noise_operator=None,
            ),
            ...,
            Sample(
                frequency=6283185.307179586,
                inverse_power=0.004755084538721049,
                inverse_power_uncertainty=6.537825035071768e-05,
                frequency_domain_noise_operator=None,
            )
        ]

    >>> qctrl.utils.extract_filter_function_arrays(samples)
        (
            array([6.28318531e-02, ..., 6.28318531e+06]),
            array([2.46904686e+00, ..., 4.75508454e-03]),
            array([1.64630559e-03, ..., 6.53782504e-05]),
            None,
        )
    """
    check_argument_iterable(samples, "samples")
    check_argument(
        all(hasattr(sample, "frequency") for sample in samples),
        "Not every sample contains a frequency.",
        {"samples": samples},
    )
    check_argument(
        all(hasattr(sample, "inverse_power") for sample in samples),
        "Not every sample contains an inverse power.",
        {"samples": samples},
    )
    check_argument(
        all(hasattr(sample, "inverse_power_uncertainty") for sample in samples),
        "Not every sample contains an inverse power uncertainty.",
        {"samples": samples},
    )
    check_argument(
        all(hasattr(sample, "frequency_domain_noise_operator") for sample in samples),
        "Not every sample contains a frequency domain noise operator.",
        {"samples": samples},
    )

    frequencies = [sample.frequency for sample in samples]
    inverse_powers = [sample.inverse_power for sample in samples]
    inverse_power_uncertainties = [
        sample.inverse_power_uncertainty for sample in samples
    ]
    frequency_domain_noise_operators = [
        sample.frequency_domain_noise_operator for sample in samples
    ]

    frequencies = (
        None
        if any(frequency is None for frequency in frequencies)
        else np.array(frequencies)
    )
    inverse_powers = (
        None
        if any(inverse_power is None for inverse_power in inverse_powers)
        else np.array(inverse_powers)
    )
    inverse_power_uncertainties = (
        None
        if any(
            inverse_power_uncertainty is None
            for inverse_power_uncertainty in inverse_power_uncertainties
        )
        else np.array(inverse_power_uncertainties)
    )
    frequency_domain_noise_operators = (
        None
        if any(
            frequency_domain_noise_operator is None
            for frequency_domain_noise_operator in frequency_domain_noise_operators
        )
        else np.array(frequency_domain_noise_operators)
    )

    return (
        frequencies,
        inverse_powers,
        inverse_power_uncertainties,
        frequency_domain_noise_operators,
    )



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Pulse library nodes.
"""

from typing import (
    Optional,
    Union,
)

import numpy as np
from qctrlcommons.graph import Graph
from qctrlcommons.node.node_data import (
    Pwc,
    Tensor,
)
from qctrlcommons.preconditions import check_argument

from qctrltoolkit.namespace import Namespace
from qctrltoolkit.toolkit_utils import expose


def _validate_optimizable_parameter(
    graph: Graph, parameter: Union[float, Tensor], name: str
) -> Tensor:
    """
    Converts parameter into a Tensor, checks that it contains a single element
    and returns a Tensor containing that element.
    """

    parameter = graph.tensor(parameter)
    check_argument(
        np.prod(parameter.shape) == 1,
        f"If passed as a Tensor, the {name} must either "
        "be a scalar or contain a single element.",
        {name: parameter},
        extras={f"{name}.shape": parameter.shape},
    )

    return graph.reshape(parameter, [1])[0]


@expose(Namespace.PULSES)
def square_pulse_pwc(
    graph: Graph,
    duration: float,
    amplitude: Union[float, complex, Tensor],
    start_time: float = 0,
    end_time: Optional[float] = None,
    name: Optional[str] = None,
) -> Pwc:
    r"""
    Creates a `Pwc` representing a square pulse.

    The entire signal lasts from time 0 to the given duration with the
    square pulse being applied from the start time to the end time.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    duration : float
        The duration of the signal.
    amplitude : float or complex or Tensor
        The amplitude of the square pulse, :math:`A`.
        It must either be a scalar or contain a single element.
    start_time : float, optional
        The start time of the square pulse, :math:`t_0`.
        Defaults to 0.
    end_time : float, optional
        The end time of the square pulse, :math:`t_1`.
        Must be greater than the start time.
        Defaults to the value of the given duration.
    name : str, optional
        The name of the node.

    Returns
    -------
    Pwc
        The square pulse. The returned Pwc can have a maximum of three segments and a minimum
        of one, depending on the start time, end time, and duration of the pulse.

    Notes
    -----
        The square pulse is defined as

        .. math:: \mathop{\mathrm{Square}}(t) = A \theta(t-t_0) \theta(t_1-t) \; ,

        where :math:`\theta(t)` is the
        `Heaviside step function <https://en.wikipedia.org/wiki/Heaviside_step_function>`_.

    Examples
    --------

    Define a square PWC pulse.

    >>> graph.pulses.square_pulse_pwc(
    ...     duration=4.0,
    ...     amplitude=2.5,
    ...     start_time=1.0,
    ...     end_time=3.0,
    ...     name="square",
    ... )
    <Pwc: name="square", operation_name="time_concatenate_pwc", value_shape=(), batch_shape=()>
    >>> result = qctrl.functions.calculate_graph(graph=graph, output_node_names=["square"])
    >>> result.output["square"]
    [
        {'duration': 1.0, 'value': 0.0},
        {'duration': 2.0, 'value': 2.5},
        {'duration': 1.0, 'value': 0.0}
    ]

    Define a square pulse with an optimizable amplitude.

    >>> amplitude = graph.optimization_variable(
    ...     count=1, lower_bound=0, upper_bound=2.*np.pi, name="amplitude"
    ... )
    >>> graph.pulses.square_pulse_pwc(
    ...     duration=4.0, amplitude=amplitude, name="square"
    ... )
    <Pwc: name="square", operation_name="time_concatenate_pwc", value_shape=(), batch_shape=()>
    """

    check_argument(
        duration > 0.0, "The duration must be positive.", {"duration": duration}
    )

    if end_time is None:
        end_time = duration

    check_argument(
        end_time > start_time,
        "The end time must be greater than the start time.",
        {"start_time": start_time, "end_time": end_time},
    )

    amplitude = _validate_optimizable_parameter(graph, amplitude, "amplitude")

    if start_time >= duration or end_time <= 0:
        # In both of these cases the signal is always zero.
        return graph.pwc_signal(duration=duration, values=np.array([0.0]), name=name)

    pwcs = []

    if start_time > 0:
        # Add preceding step function.
        pwcs.append(graph.pwc_signal(duration=start_time, values=np.array([0.0])))

    pwcs.append(
        graph.pwc_signal(
            duration=min(end_time, duration) - max(start_time, 0),
            values=amplitude * np.array([1.0]),
        )
    )

    if end_time < duration:
        # Add trailing step function.
        pwcs.append(
            graph.pwc_signal(duration=duration - end_time, values=np.array([0.0]))
        )

    return graph.time_concatenate_pwc(pwcs, name=name)


@expose(Namespace.PULSES)
def sech_pulse_pwc(
    graph: Graph,
    duration: float,
    segment_count: int,
    amplitude: Union[float, complex, Tensor],
    width: Optional[Union[float, Tensor]] = None,
    center_time: Optional[Union[float, Tensor]] = None,
    name: Optional[str] = None,
) -> Pwc:
    r"""
    Creates a `Pwc` representing a hyperbolic secant pulse.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    duration : float
        The duration of the signal, :math:`T`.
    segment_count : int
        The number of segments in the PWC.
    amplitude : float or complex or Tensor
        The amplitude of the pulse, :math:`A`.
        It must either be a scalar or contain a single element.
    width : float or Tensor, optional
        The characteristic time for the hyperbolic secant pulse, :math:`t_\mathrm{pulse}`.
        If passed, it must either be a scalar or contain a single element.
        Defaults to :math:`T/12`,
        giving the pulse a full width at half maximum (FWHM) of :math:`0.22 T`.
    center_time : float or Tensor, optional
        The time at which the pulse peaks, :math:`t_\mathrm{peak}`.
        If passed, it must either be a scalar or contain a single element.
        Defaults to :math:`T/2`.
    name : str, optional
        The name of the node.

    Returns
    -------
    Pwc
        The hyperbolic secant pulse.

    Notes
    -----
    The hyperbolic secant pulse is defined as

        .. math:: \mathop{\mathrm{Sech}}(t)
            = \frac{A}{\cosh\left((t - t_\mathrm{peak}) / t_\mathrm{pulse} \right)} \;.

    The FWHM of the pulse is about :math:`2.634 t_\mathrm{pulse}`.

    Examples
    --------

    Define a simple sech PWC pulse.

    >>> graph.pulses.sech_pulse_pwc(
    ...     duration=5, segment_count=50, amplitude=1, name="sech_pulse"
    ... )
    <Pwc: name="sech_pulse", operation_name="discretize_stf", value_shape=(), batch_shape=()>
    >>> result = qctrl.functions.calculate_graph(graph=graph, output_node_names=["sech_pulse"])
    >>> result.output["sech_pulse"]
    [
        {'value': -0.0056, 'duration': 0.1},
        {'value': -0.0071, 'duration': 0.1},
        ...
        {'value': 0.0071, 'duration': 0.1},
        {'value': 0.0056, 'duration': 0.1},
    ]

    Define a displaced sech PWC pulse.

    >>> graph.pulses.sech_pulse_pwc(
    ...     duration=3e-6,
    ...     segment_count=60,
    ...     amplitude=20e6,
    ...     width=0.15e-6,
    ...     center_time=1e-6,
    ...     name="sech_pulse",
    ... )
    <Pwc: name="sech_pulse", operation_name="discretize_stf", value_shape=(), batch_shape=()>
    >>> result = qctrl.functions.calculate_graph(graph=graph, output_node_names=["sech_pulse"])
    >>> result.output["sech_pulse"]
    [
        {'value': 60137.43, 'duration': 5.e-08},
        {'value': 83928.37, 'duration': 5.e-08},
        ...
        {'value': 106.8105, 'duration': 5.e-08},
        {'value': 76.53310, 'duration': 5.e-08},
    ]

    Define a sech pulse with optimizable parameters.

    >>> graph = qctrl.create_graph()
    >>> amplitude = graph.optimization_variable(
    ...     count=1, lower_bound=0, upper_bound=10e6, name="amplitude"
    ... )
    >>> width = graph.optimization_variable(
    ...     count=1, lower_bound=0.1e-6, upper_bound=0.5e-6, name="width"
    ... )
    >>> center_time = graph.optimization_variable(
    ...     count=1, lower_bound=1e-6, upper_bound=2e-6, name="center_time"
    ... )
    >>> graph.pulses.sech_pulse_pwc(
    ...     duration=3e-6,
    ...     segment_count=32,
    ...     amplitude=amplitude,
    ...     width=width,
    ...     center_time=center_time,
    ...     name="sech_pulse",
    ... )
    <Pwc: name="sech_pulse", operation_name="discretize_stf", value_shape=(), batch_shape=()>
    """

    check_argument(
        duration > 0.0, "The duration must be positive.", {"duration": duration}
    )

    if width is None:
        width = duration / 12

    if center_time is None:
        center_time = duration / 2

    amplitude = _validate_optimizable_parameter(graph, amplitude, "amplitude")
    width = _validate_optimizable_parameter(graph, width, "width")
    center_time = _validate_optimizable_parameter(graph, center_time, "center time")

    stf = amplitude / graph.cosh((graph.identity_stf() - center_time) / width)
    return graph.discretize_stf(stf, duration, segment_count, name=name)


@expose(Namespace.PULSES)
def linear_ramp_pwc(
    graph: Graph,
    duration: float,
    segment_count: int,
    end_value: Union[float, complex, Tensor],
    start_value: Optional[Union[float, complex, Tensor]] = None,
    start_time: float = 0.0,
    end_time: Optional[float] = None,
    name: Optional[str] = None,
) -> Pwc:
    r"""
    Creates a `Pwc` representing a linear ramp.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    duration : float
        The duration of the signal, :math:`T`.
    segment_count : int
        The number of segments in the PWC.
    end_value : float or complex or Tensor
        The value of the ramp at :math:`t = t_\mathrm{end}`, :math:`a_\mathrm{end}`.
        It must either be a scalar or contain a single element.
    start_value : float or complex or Tensor, optional
        The value of the ramp at :math:`t = t_\mathrm{start}`, :math:`a_\mathrm{start}`.
        If passed, it must either be a scalar or contain a single element.
        Defaults to :math:`-a_\mathrm{end}`.
    start_time : float, optional
        The time at which the linear ramp starts, :math:`t_\mathrm{start}`.
        Defaults to zero.
    end_time : float, optional
        The time at which the linear ramp ends, :math:`t_\mathrm{end}`.
        Defaults to the given duration :math:`T`.
    name : str, optional
        The name of the node.

    Returns
    -------
    Pwc
        The linear ramp.

    Notes
    -----
        The linear ramp is defined as

        .. math:: \mathop{\mathrm{Linear}}(t) =
            \begin{cases}
                a_\mathrm{start} &\mathrm{if} \quad t < t_\mathrm{start}\\
                a_\mathrm{start} + (a_\mathrm{end} - a_\mathrm{start})
                    \frac{t - t_\mathrm{start}}{t_\mathrm{end} - t_\mathrm{start}}
                    &\mathrm{if} \quad t_\mathrm{start} \le t \le t_\mathrm{end} \\
                a_\mathrm{end} &\mathrm{if} \quad t > t_\mathrm{end}
            \end{cases} \;.

    Examples
    --------

    Define a linear PWC ramp.

    >>> graph.pulses.linear_ramp_pwc(
    ...     duration=2.0, segment_count=5, end_value=1.5, start_value=0.5, name="linear_ramp"
    ... )
    <Pwc: name="linear_ramp", operation_name="discretize_stf", value_shape=(), batch_shape=()>
    >>> result = qctrl.functions.calculate_graph(graph=graph, output_node_names=["linear_ramp"])
    >>> result.output["linear_ramp"]
    [
        {'duration': 0.4, 'value': 0.6},
        {'duration': 0.4, 'value': 0.8},
        {'duration': 0.4, 'value': 1.0},
        {'duration': 0.4, 'value': 1.2},
        {'duration': 0.4, 'value': 1.4},
    ]

    Define a linear ramp with start and end times.

    >>> graph.pulses.linear_ramp_pwc(
    ...     duration=4, segment_count=5, end_value=2, start_time=1, end_time=3, name="linear_ramp"
    ... )
    <Pwc: name="linear_ramp", operation_name="time_concatenate_pwc", value_shape=(), batch_shape=()>
    >>> result = qctrl.functions.calculate_graph(graph=graph, output_node_names=["linear_ramp"])
    >>> result.output["linear_ramp"]
    [
        {'value': -2.0, 'duration': 1.0},
        {'value': -1.5, 'duration': 0.5},
        {'value': -0.5, 'duration': 0.5},
        {'value': 0.5, 'duration': 0.5},
        {'value': 1.5, 'duration': 0.5},
        {'value': 2.0, 'duration': 1.0},
    ]

    Define a linear ramp with an optimizable slope around 0.

    >>> duration = 4.0
    >>> slope = graph.optimization_variable(
    ...     count=1, lower_bound=-30, upper_bound=30, name="slope"
    ... )
    >>> end_value = slope * duration / 2
    >>> graph.pulses.linear_ramp_pwc(
    ...     duration=duration, segment_count=64, end_value=end_value, name="linear_ramp"
    ... )
    <Pwc: name="linear_ramp", operation_name="discretize_stf", value_shape=(), batch_shape=()>
    """

    check_argument(
        duration > 0.0, "The duration must be positive.", {"duration": duration}
    )

    if start_value is None:
        start_value = -end_value

    if end_time is None:
        end_time = duration

    check_argument(
        start_time < end_time,
        "The end time of the pulse must be greater than the start time.",
        {"end_time": end_time, "start_time": start_time},
    )

    end_value = _validate_optimizable_parameter(graph, end_value, "end value")
    start_value = _validate_optimizable_parameter(graph, start_value, "start value")

    slope = (end_value - start_value) / (end_time - start_time)

    if start_time <= 0.0 and end_time >= duration:
        # No flat parts inside of the PWC.
        stf = start_value + slope * (graph.identity_stf() - start_time)
        return graph.discretize_stf(stf, duration, segment_count, name=name)

    if end_time <= 0.0 or start_time >= duration:
        # The whole ramp falls outside of the PWC.
        return graph.discretize_stf(
            graph.constant_stf(0.0), duration, segment_count, name=name
        )

    pre_step_pwc = None
    post_step_pwc = None

    if start_time > 0.0:
        # Preceding step function.
        pre_step_pwc = graph.constant_pwc(constant=start_value, duration=start_time)
        segment_count -= 1

    if end_time < duration:
        # Trailing step function.
        post_step_pwc = graph.constant_pwc(
            constant=end_value, duration=duration - end_time
        )
        segment_count -= 1

    # Ramp part.
    stf = start_value + slope * (graph.identity_stf() - min(start_time, 0.0))
    ramp_duration = min(end_time, duration) - max(start_time, 0.0)
    ramp_pwc = graph.discretize_stf(stf, ramp_duration, segment_count)

    pwcs = [ramp_pwc]
    if pre_step_pwc is not None:
        pwcs.insert(0, pre_step_pwc)
    if post_step_pwc is not None:
        pwcs.append(post_step_pwc)

    return graph.time_concatenate_pwc(pwcs, name=name)


@expose(Namespace.PULSES)
def tanh_ramp_pwc(
    graph: Graph,
    duration: float,
    segment_count: int,
    end_value: Union[float, complex, Tensor],
    start_value: Optional[Union[float, complex, Tensor]] = None,
    ramp_duration: Optional[Union[float, Tensor]] = None,
    center_time: Optional[Union[float, Tensor]] = None,
    name: Optional[str] = None,
) -> Pwc:
    r"""
    Creates a `Pwc` representing a hyperbolic tangent ramp.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    duration : float
        The duration of the signal, :math:`T`.
    segment_count : int
        The number of segments in the PWC.
    end_value : float or complex or Tensor
        The asymptotic value of the ramp towards :math:`t \to +\infty`, :math:`a_+`.
        It must either be a scalar or contain a single element.
    start_value : float or complex or Tensor, optional
        The asymptotic value of the ramp towards :math:`t \to -\infty`, :math:`a_-`.
        If passed, it must either be a scalar or contain a single element.
        Defaults to minus `end_value`.
    ramp_duration : float or Tensor, optional
        The characteristic time for the hyperbolic tangent ramp, :math:`t_\mathrm{ramp}`.
        If passed, it must either be a scalar or contain a single element.
        Defaults to :math:`T/6`.
    center_time : float or Tensor, optional
        The time at which the ramp has its greatest slope, :math:`t_0`.
        If passed, it must either be a scalar or contain a single element.
        Defaults to :math:`T/2`.
    name : str, optional
        The name of the node.

    Returns
    -------
    Pwc
        The hyperbolic tangent ramp.

    Notes
    -----
    The hyperbolic tangent ramp is defined as

        .. math:: \mathop{\mathrm{Tanh}}(t)
            = \frac{a_+ + a_-}{2}
                + \frac{a_+ - a_-}{2} \tanh\left( \frac{t - t_0}{t_\mathrm{ramp}} \right) \; ,

    where the function's asymptotic values :math:`a_\pm` are defined by:

        .. math::  a_\pm := \lim_{t\to\pm\infty} \mathop{\mathrm{Tanh}}(t) \;,

    and :math:`t_0` is related to :math:`t_\mathrm{ramp}` by:

        .. math::
            \left.\frac{{\rm d}\mathop{\mathrm{Tanh}}(t)}{{\rm d}t}\right|_{t=t_0}
                = \frac{ (a_+ - a_-)}{2 t_\mathrm{ramp}} \; .

    Note that if :math:`t_0` is close to the edges of the PWC,
    for example :math:`t_0 \lesssim 2 t_\mathrm{ramp}`,
    then the first and last values of the PWC will differ from the expected asymptotic values.

    With the default values of `start_value` (:math:`a_-`),
    `ramp_duration` (:math:`t_\mathrm{ramp}`), and `center_time` (:math:`t_0`),
    the ramp expression simplifies to

        .. math:: \mathop{\mathrm{Tanh}}(t) = A \tanh\left( \frac{t - T/2}{T/6} \right)\; ,

    where :math:`A = a_+` is the end value (the start value is then :math:`-A`).
    This defines a symmetric ramp (around :math:`(T/2, 0)`)
    between :math:`-0.995 A` (at :math:`t=0`) and :math:`0.995 A` (at :math:`t=T`).

    Examples
    --------

    Define a simple tanh PWC ramp.

    >>> graph.pulses.tanh_ramp_pwc(
    ...     duration=5.0, segment_count=50, end_value=1, name="tanh_ramp"
    ... )
    <Pwc: name="tanh_ramp", operation_name="discretize_stf", value_shape=(), batch_shape=()>
    >>> result = qctrl.functions.calculate_graph(graph=graph, output_node_names=["tanh_ramp"])
    >>> result.output["tanh_ramp"]
    [
        {'value': -0.9944, 'duration': 0.1},
        {'value': -0.9929, 'duration': 0.1},
        ...
        {'value': 0.9929, 'duration': 0.1},
        {'value': 0.9944, 'duration': 0.1},
    ]

    Define a flat-top pulse from two tanh ramps.

    >>> ramp = graph.pulses.tanh_ramp_pwc(
    ...     duration=3,
    ...     segment_count=60,
    ...     end_value=1,
    ...     ramp_duration=0.25,
    ...     center_time=0.5,
    ... )
    >>> flat_top_pulse = 0.5 * (ramp + graph.time_reverse_pwc(ramp))
    >>> flat_top_pulse.name="flat_top_pulse"
    >>> print(flat_top_pulse)
    <Pwc: name="flat_top_pulse", operation_name="multiply", value_shape=(), batch_shape=()>
    >>> result = qctrl.functions.calculate_graph(graph=graph, output_node_names=["flat_top_pulse"])
    >>> result.output["flat_top_pulse"]
    [
        {'value': 0.0219, 'duration': 0.05},
        {'value': 0.0323, 'duration': 0.05},
        ...
        {'value': 0.0323, 'duration': 0.05},
        {'value': 0.0219, 'duration': 0.05},
    ]

    Define a tanh ramp with optimizable parameters.

    >>> end_value = graph.optimization_variable(
    ...     count=1, lower_bound=0, upper_bound=3e6, name="end_value"
    ... )
    >>> ramp_duration = graph.optimization_variable(
    ...     count=1, lower_bound=0.1e-6, upper_bound=0.3e-6, name="ramp_duration"
    ... )
    >>> center_time = graph.optimization_variable(
    ...     count=1, lower_bound=0.25e-6, upper_bound=0.75e-6, name="center_time"
    ... )
    >>> graph.pulses.tanh_ramp_pwc(
    ...     duration=1e-6,
    ...     segment_count=32,
    ...     end_value=end_value,
    ...     start_value=0.0,
    ...     ramp_duration=ramp_duration,
    ...     center_time=center_time,
    ...     name="tanh_ramp",
    ... )
    <Pwc: name="tanh_ramp", operation_name="discretize_stf", value_shape=(), batch_shape=()>
    """

    check_argument(
        duration > 0.0, "The duration must be positive.", {"duration": duration}
    )

    if start_value is None:
        start_value = -end_value

    if ramp_duration is None:
        ramp_duration = duration / 6

    if center_time is None:
        center_time = duration / 2

    end_value = _validate_optimizable_parameter(graph, end_value, "end value")
    start_value = _validate_optimizable_parameter(graph, start_value, "start value")
    ramp_duration = _validate_optimizable_parameter(
        graph, ramp_duration, "ramp duration"
    )
    center_time = _validate_optimizable_parameter(graph, center_time, "center time")

    stf = start_value + 0.5 * (end_value - start_value) * (
        1 + graph.tanh((graph.identity_stf() - center_time) / ramp_duration)
    )
    return graph.discretize_stf(stf, duration, segment_count, name=name)


def _allocate_segment_counts(durations: list, segment_count: int):
    """
    Allocates the number of segments between two non-flat segments and one flat segment.
    The flat segment is assumed to lie in the middle of the other two.
    The number of segments each non-flat part has is in proportion to their duration of time.
    The non-flat durations are at indexes 0 and 2 of durations, durations[0] and durations[2].
    durations[1] can only be zero if one of durations[0] or durations[2] are also zero
    (e.g. [2,0,3] and [0,0,0] are not valid inputs).
    """
    # In the symmetric case (durations[0] = durations[2]), the flat part gets two segments
    # if segment_count is even and one otherwise.
    if np.isclose(durations[0], durations[2]):
        pulse_counts = [(segment_count - 1) // 2] * 2
        pulse_counts.insert(1, segment_count - 2 * pulse_counts[0])
        return pulse_counts

    flat_segments = int(durations[1] > 0.0)  # Number of flat segments.
    # Note that if flat_segments is 0, either durations[0] or durations[2] must be 0.
    if np.isclose(durations[0], 0.0):
        return [0, flat_segments, segment_count - flat_segments]
    if np.isclose(durations[2], 0.0):
        return [segment_count - flat_segments, flat_segments, 0]

    # In this case we know flat_segments=1 and all pulse_counts should be at least 1.
    non_flat_duration = durations[0] + durations[2]
    pulse_counts = [
        int((segment_count - 1) * durations[index] / non_flat_duration)
        for index in [0, 2]
    ]
    # Make sure the pulses each get at least one segment even if the duration is very small.
    pulse_counts = [max(pulse_count, 1) for pulse_count in pulse_counts]
    pulse_counts.insert(1, 1)

    # If there are too many or few segments, we update the largest segment count.
    pulse_counts[np.argmax(pulse_counts)] += segment_count - sum(pulse_counts)

    return pulse_counts


@expose(Namespace.PULSES)
def gaussian_pulse_pwc(
    graph: Graph,
    duration: float,
    segment_count: int,
    amplitude: Union[float, complex, Tensor],
    width: Optional[Union[float, Tensor]] = None,
    center_time: Optional[float] = None,
    drag: Optional[Union[float, Tensor]] = None,
    flat_duration: Optional[float] = None,
    name: Optional[str] = None,
) -> Pwc:
    r"""
    Creates a `Pwc` representing a Gaussian pulse.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    duration : float
        The duration of the signal, :math:`T`.
    segment_count : int
        The number of segments in the PWC.
        Must be at least four.
    amplitude : float or complex or Tensor
        The amplitude of the Gaussian pulse, :math:`A`.
        It must either be a scalar or contain a single element.
    width : float or Tensor, optional
        The standard deviation of the Gaussian pulse, :math:`\sigma`.
        It must either be a scalar or contain a single element.
        Defaults to :math:`T/10` or :math:`(T-t_\mathrm{flat})/10` if `flat_duration` is passed.
    center_time : float, optional
        The center of the Gaussian pulse, :math:`t_0`.
        Defaults to half of the given value of the duration, :math:`T/2`.
    drag : float or Tensor, optional
        The DRAG parameter, :math:`\beta`.
        If passed, it must either be a scalar or contain a single element.
        Defaults to no DRAG correction.
    flat_duration : float, optional
        The amount of time to remain constant after the peak of the Gaussian,
        :math:`t_\mathrm{flat}`.
        If passed, it must be positive and less than the duration.
        Defaults to None, in which case no constant part is added to the Gaussian pulse.
    name : str, optional
        The name of the node.

    Returns
    -------
    Pwc
        The sampled Gaussian pulse.
        If no flat duration is passed then the pulse is evenly sampled between :math:`0` and
        :math:`T`. If one is passed, the flat part of the pulse is described by one or two segments
        (depending on the values of `center_time` and `segment_count`), and the rest of the pulse is
        sampled with the remaining segments.

    Notes
    -----
        The Gaussian pulse is defined as

        .. math:: \mathop{\mathrm{Gaussian}}(t) =
           \begin{cases}
                A \left(1-\frac{i\beta (t-t_1)}{\sigma^2}\right)
                \exp \left(- \frac{(t-t_1)^2}{2\sigma^2} \right)
                    &\mathrm{if} \quad t < t_1=t_0- t_\mathrm{flat}/2\\
                A
                    &\mathrm{if} \quad t_0-t_\mathrm{flat}/2 \le t < t_0+t_\mathrm{flat}/2 \\
                A \left(1-\frac{i\beta (t-t_2)}{\sigma^2}\right)
                \exp \left(- \frac{(t-t_2)^2}{2\sigma^2} \right)
                    &\mathrm{if} \quad t > t_2=t_0+t_\mathrm{flat}/2
            \end{cases} \;.

        If the flat duration is zero (the default setting), this reduces to

        .. math:: \mathop{\mathrm{Gaussian}}(t) =
            A \left(1-\frac{i\beta (t-t_0)}{\sigma^2}\right)
            \exp \left(- \frac{(t-t_0)^2}{2\sigma^2} \right) \; .

    Examples
    --------

    Define a Gaussian PWC pulse.

    >>> graph.pulses.gaussian_pulse_pwc(
    ...     duration=3.0,
    ...     segment_count=100,
    ...     amplitude=1.0,
    ...     name="gaussian",
    ... )
    <Pwc: name="gaussian", operation_name="discretize_stf", value_shape=(), batch_shape=()>
    >>> result = qctrl.functions.calculate_graph(graph=graph, output_node_names=["gaussian"])
    >>> result.output["gaussian"]
    [
        {'duration': 0.03, 'value': 4.7791e-06},
        {'duration': 0.03, 'value': 7.8010e-06},
        ...
        {'duration': 0.03, 'value': 7.8010e-06},
        {'duration': 0.03, 'value': 4.7791e-06}
    ]

    Define a flat-top Gaussian PWC pulse with a DRAG correction.

    >>> graph.pulses.gaussian_pulse_pwc(
    ...     duration=3.0,
    ...     segment_count=100,
    ...     amplitude=1.0,
    ...     width=0.2,
    ...     center_time=1.5,
    ...     drag=0.1,
    ...     flat_duration=0.2,
    ...     name="gaussian",
    ... )
    <Pwc: name="gaussian", operation_name="time_concatenate_pwc", value_shape=(), batch_shape=()>
    >>> result = qctrl.functions.calculate_graph(graph=graph, output_node_names=["gaussian"])
    >>> result.output["gaussian"]
    [
        {'duration': 0.0285, 'value': (3.7655e-11+1.3044e-10j)},
        {'duration': 0.0285, 'value': (1.0028e-10+3.4026e-10j)},
        ...
        {'duration': 0.0285, 'value': (1.0028e-10-3.4026e-10j)},
        {'duration': 0.0285, 'value': (3.7655e-11-1.3044e-10j)}
    ]

    Define a Gaussian pulse with optimizable parameters.

    >>> amplitude = graph.optimization_variable(
    ...     count=1, lower_bound=0, upper_bound=2.*np.pi, name="amplitude"
    ... )
    >>> width = graph.optimization_variable(
    ...     count=1, lower_bound=0, upper_bound=2., name="width"
    ... )
    >>> drag = graph.optimization_variable(
    ...     count=1, lower_bound=0, upper_bound=1., name="drag"
    ... )
    >>> graph.pulses.gaussian_pulse_pwc(
    ...     duration=3.0,
    ...     segment_count=100,
    ...     amplitude=amplitude,
    ...     width=width,
    ...     drag=drag,
    ...     name="gaussian",
    ... )
    <Pwc: name="gaussian", operation_name="discretize_stf", value_shape=(), batch_shape=()>
    """

    check_argument(
        duration > 0.0, "The duration must be positive.", {"duration": duration}
    )

    if center_time is None:
        center_time = 0.5 * duration

    if width is None:
        if flat_duration is None:
            width = duration / 10
        else:
            width = (duration - flat_duration) / 10

    amplitude = _validate_optimizable_parameter(graph, amplitude, "amplitude")
    width = _validate_optimizable_parameter(graph, width, "width")

    def create_gaussian(center_parameter):
        return graph.exp(
            -((graph.identity_stf() - center_parameter) ** 2) / (2 * width**2)
        )

    def create_amplitude(center_parameter):
        if drag is None:
            return amplitude
        return amplitude * (
            1.0 - 1j * drag * (graph.identity_stf() - center_parameter) / (width**2)
        )

    if drag is not None:
        drag = _validate_optimizable_parameter(graph, drag, "drag")

    if flat_duration is None:
        return graph.discretize_stf(
            create_amplitude(center_time) * create_gaussian(center_time),
            duration,
            segment_count,
            name=name,
        )

    check_argument(
        0.0 < flat_duration < duration,
        "The flat duration must be positive and less than the duration.",
        {"flat_duration": flat_duration},
        extras={"duration": duration},
    )

    check_argument(
        segment_count > 3,
        "The number of segments must be at least 4.",
        {"segment_count": segment_count},
    )

    # Time at which first Gaussian ends.
    flat_duration_start = center_time - 0.5 * flat_duration
    # Time at which second Gaussian starts.
    flat_duration_end = center_time + 0.5 * flat_duration

    if flat_duration_start >= duration:
        # In this case since the flat segment starts after the duration, it is not part of
        # the signal.
        return graph.discretize_stf(
            create_amplitude(flat_duration_start)
            * create_gaussian(flat_duration_start),
            duration,
            segment_count,
            name=name,
        )

    if flat_duration_end <= 0:
        # In this case since the flat segment finishes before time zero, it is not part of
        # the signal.
        return graph.discretize_stf(
            create_amplitude(flat_duration_end) * create_gaussian(flat_duration_end),
            duration,
            segment_count,
            name=name,
        )

    if flat_duration_end >= duration:
        # In this case since the flat segment finishes after the duration, the second Gaussian
        # is not part of the signal.
        durations = [flat_duration_start, duration - flat_duration_start]
        segment_counts = [segment_count - 1, 1]
        stfs = [
            create_amplitude(flat_duration_start)
            * create_gaussian(flat_duration_start),
            graph.constant_stf(amplitude),
        ]

    elif flat_duration_start <= 0:
        # In this case since the flat segment begins before time zero, the first Gaussian
        # is not part of the signal.
        durations = [flat_duration_end, duration - flat_duration_end]
        segment_counts = [1, segment_count - 1]
        stfs = [
            graph.constant_stf(amplitude),
            create_amplitude(0.0) * create_gaussian(0.0),
        ]

    else:
        durations = [flat_duration_start, flat_duration, duration - flat_duration_end]
        segment_counts = _allocate_segment_counts(durations, segment_count)
        stfs = [
            create_amplitude(flat_duration_start)
            * create_gaussian(flat_duration_start),
            graph.constant_stf(amplitude),
            create_amplitude(0.0) * create_gaussian(0.0),
        ]

    pwcs = [
        graph.discretize_stf(*args) for args in zip(stfs, durations, segment_counts)
    ]
    return graph.time_concatenate_pwc(pwcs, name=name)


@expose(Namespace.PULSES)
def cosine_pulse_pwc(
    graph: Graph,
    duration: float,
    segment_count: int,
    amplitude: Union[float, complex, Tensor],
    drag: Optional[Union[float, Tensor]] = None,
    start_time: float = 0.0,
    end_time: Optional[float] = None,
    flat_duration: Optional[float] = None,
    name: Optional[str] = None,
) -> Pwc:
    r"""
    Creates a `Pwc` representing a cosine pulse.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    duration : float
        The duration of the signal, :math:`T`.
    segment_count : int
        The number of segments in the PWC.
        Must be at least six.
    amplitude : float or complex or Tensor
        The amplitude of the pulse, :math:`A`.
        It must either be a scalar or contain a single element.
    drag : float or Tensor, optional
        The DRAG parameter, :math:`\beta`.
        If passed, it must either be a scalar or contain a single element.
        Defaults to no DRAG correction.
    start_time : float, optional
        The time at which the cosine pulse starts, :math:`t_\mathrm{start}`.
        Defaults to zero.
    end_time : float, optional
        The time at which the cosine pulse ends, :math:`t_\mathrm{end}`.
        Defaults to the given duration :math:`T`.
    flat_duration : float, optional
        The amount of time to remain constant after the peak of the cosine,
        :math:`t_\mathrm{flat}`.
        If passed, it must be positive and less than the difference between `end_time` and
        `start_time`.
        Defaults to None, in which case no constant part is added to the cosine pulse.
    name : str, optional
        The name of the node.

    Returns
    -------
    Pwc
        The sampled cosine pulse.
        If no flat duration is passed then the pulse is evenly sampled between :math:`0` and
        :math:`T`. If one is passed, the flat part of the pulse is described by one or two segments
        (depending on the value of `segment_count`), and the rest of the pulse is sampled with the
        remaining segments.

    See Also
    --------
    :func:`.pulses.hann_series_pwc` :
        Creates a `Pwc` representing a sum of Hann window functions.
    :func:`.pulses.sinusoid_pwc` :
        Creates a `Pwc` representing a sinusoidal oscillation.

    Notes
    -----
        The cosine pulse is defined as

        .. math:: \mathop{\mathrm{Cos}}(t) =
            \begin{cases}
            0
            &\mathrm{if} \quad t < t_\mathrm{start} \\
            \frac{A}{2} \left[1+\cos \left(\omega \{t-\tau_-\} \right)
            + i\omega\beta \sin \left(\omega \{t-\tau_-\}\right)\right]
            &\mathrm{if} \quad t_\mathrm{start} \le t < \tau_- \\
            A
            &\mathrm{if} \quad \tau_- \le t \le \tau_+ \\
            \frac{A}{2} \left[1+\cos \left(\omega\{t-\tau_+\}\right)
            + i\omega \beta\sin \left(\omega \{t-\tau_+\}\right)\right]
            &\mathrm{if} \quad \tau_+ < t \le t_\mathrm{end} \\
            0
            &\mathrm{if} \quad t > t_\mathrm{end} \\
            \end{cases}\; ,

        where :math:`\omega=2\pi /(t_\mathrm{end}-t_\mathrm{start} - t_\mathrm{flat})`,
        :math:`\tau_\mp` are the start/end times of the flat segment,
        with :math:`\tau_\mp=(t_\mathrm{start}+t_\mathrm{end} \mp t_\mathrm{flat})/2`.

        If the flat duration is zero (the default setting), this reduces to

        .. math:: \mathop{\mathrm{Cos}}(t) =
            \frac{A}{2} \left[1+\cos \left(\omega \{t-\tau\} \right)
            + i\omega\beta \sin \left(\omega \{t-\tau\}\right)\right]
            \theta(t-t_\mathrm{start}) \theta(t_\mathrm{end}-t)\; ,

        where now :math:`\omega=2\pi /(t_\mathrm{end}-t_\mathrm{start})`,
        :math:`\tau=(t_\mathrm{start}+t_\mathrm{end})/2`
        and :math:`\theta(t)` is the
        `Heaviside step function <https://en.wikipedia.org/wiki/Heaviside_step_function>`_.

    Examples
    --------

    Define a cosine PWC pulse.

    >>> graph.pulses.cosine_pulse_pwc(
    ...     duration=3.0, segment_count=100, amplitude=1.0, name="cos_pulse"
    ... )
    <Pwc: name="cos_pulse", operation_name="time_concatenate_pwc", value_shape=(), batch_shape=()>
    >>> result = qctrl.functions.calculate_graph(graph=graph, output_node_names=["cos_pulse"])
    >>> result.output["cos_pulse"]
    [
        {'duration': 0.03, 'value': 0.00024},
        {'duration': 0.03, 'value': 0.00221},
        ...
        {'duration': 0.03, 'value': 0.00221},
        {'duration': 0.03, 'value': 0.00024}
    ]

    Define a flat-top cosine PWC pulse with a DRAG correction.

    >>> graph.pulses.cosine_pulse_pwc(
    ...     duration=3.0,
    ...     segment_count=100,
    ...     amplitude=1.0,
    ...     drag=0.1,
    ...     start_time=1.0,
    ...     end_time=2.0,
    ...     flat_duration=0.3,
    ...     name="cos_pulse",
    ... )
    <Pwc: name="cos_pulse", operation_name="time_concatenate_pwc", value_shape=(), batch_shape=()>
    >>> result = qctrl.functions.calculate_graph(graph=graph, output_node_names=["cos_pulse"])
    >>> result.output["cos_pulse"]
    [
        {'duration': 1.0, 'value': 0j},
        {'duration': 0.0072, 'value': (0.0002-0.0146j)}
        ...
        {'duration': 0.0072, 'value': (0.0002+0.0146j)},
        {'duration': 1.0, 'value': 0j}
    ]

    Define a cosine pulse with optimizable parameters.

    >>> amplitude = graph.optimization_variable(
    ...     count=1, lower_bound=0, upper_bound=2.*np.pi, name="amplitude"
    ... )
    >>> drag = graph.optimization_variable(
    ...     count=1, lower_bound=0, upper_bound=1., name="drag"
    ... )
    >>> graph.pulses.cosine_pulse_pwc(
    ...     duration=3.0,
    ...     segment_count=100,
    ...     amplitude=amplitude,
    ...     drag=drag,
    ...     name="cos_pulse",
    ... )
    <Pwc: name="cos_pulse", operation_name="time_concatenate_pwc", value_shape=(), batch_shape=()>
    """

    check_argument(
        duration > 0.0, "The duration must be positive.", {"duration": duration}
    )

    if end_time is None:
        end_time = duration

    check_argument(
        start_time < end_time,
        "The end time of the pulse must be greater than the start time.",
        {"end_time": end_time, "start_time": start_time},
    )

    if start_time >= duration or end_time <= 0:
        # In both of these cases the signal is always zero.
        return graph.pwc_signal(
            duration=duration, values=np.zeros(segment_count), name=name
        )

    amplitude = _validate_optimizable_parameter(graph, amplitude, "amplitude")

    if drag is not None:
        drag = _validate_optimizable_parameter(graph, drag, "drag")

    def create_pulse(
        peak_parameter, omega_parameter, duration_parameter, segment_count_parameter
    ):
        # Samples the function
        #   A/2[1 + \cos(\omega\{t-t_p\}) + i\omega \beta \sin(\omega\{t - t_p\})]
        # between 0 and duration_parameter,
        # where A is the amplitude, \omega the omega_parameter, t_p the peak_parameter
        # and \beta the DRAG parameter.
        # segment_count_parameter is the number of the returned PWC.

        times, time_step = np.linspace(
            0, duration_parameter, segment_count_parameter, endpoint=False, retstep=True
        )

        shifted_times = (times + time_step / 2) - peak_parameter

        pulse = 1 + np.cos(omega_parameter * shifted_times)
        if drag is not None:
            pulse = pulse + omega_parameter * drag * 1j * np.sin(
                omega_parameter * shifted_times
            )

        return graph.pwc_signal(
            duration=duration_parameter, values=0.5 * amplitude * pulse
        )

    pwcs = []

    pulse_segment_count = (
        segment_count - (start_time > 0.0) - (end_time < duration)
    )  # The number of segments for the cosine pulse.
    pulse_period = end_time - start_time  # The period of the cosine pulse.

    if start_time > 0.0:
        # Add preceding step function.
        pwcs.append(graph.pwc_signal(duration=start_time, values=np.array([0.0])))

    if flat_duration is None:
        angular_freq = 2 * np.pi / pulse_period
        # The peak is 0.5 * pulse_period after the
        # start of the pulse plus start_time if start_time < 0.
        pwcs.append(
            create_pulse(
                0.5 * pulse_period + min(start_time, 0),
                angular_freq,
                min(duration, end_time) - max(start_time, 0),
                pulse_segment_count,
            )
        )
    else:
        check_argument(
            0 < flat_duration < pulse_period,
            "The flat duration must be positive and less than the end time of the pulse minus the "
            "start time.",
            {
                "flat_duration": flat_duration,
                "start_time": start_time,
                "end_time": end_time,
            },
            extras={"end_time - start_time": end_time - start_time},
        )
        check_argument(
            segment_count > 5,
            "The number of segments must be at least 6.",
            {"segment_count": segment_count},
        )

        pulse_period -= flat_duration
        angular_freq = 2 * np.pi / pulse_period
        pulse_center = 0.5 * (start_time + end_time)
        flat_segment_start = pulse_center - flat_duration * 0.5
        flat_segment_end = pulse_center + flat_duration * 0.5

        durations = np.array(
            [
                min(flat_segment_start, duration) - max(start_time, 0),
                min(flat_segment_end, duration) - max(flat_segment_start, 0),
                min(end_time, duration) - max(flat_segment_end, 0),
            ]
        )
        durations = durations * (durations > 0.0)

        if flat_segment_start <= 0.0 and flat_segment_end >= duration:
            # In this case the flat segment is the entire signal.
            segment_counts = [0, pulse_segment_count, 0]
        else:
            segment_counts = _allocate_segment_counts(durations, pulse_segment_count)

        if flat_segment_start > 0.0:
            pwcs.append(
                create_pulse(
                    0.5 * pulse_period + min(start_time, 0),
                    angular_freq,
                    durations[0],
                    segment_counts[0],
                )
            )

        if flat_segment_start < duration and flat_segment_end > 0.0:
            pwcs.append(
                graph.discretize_stf(
                    graph.constant_stf(amplitude), durations[1], segment_counts[1]
                )
            )

        if flat_segment_end < duration:
            # The peak is at the start of the pulse, unless flat_segment_end < 0 in which case it's
            # flat_segment_end.
            pwcs.append(
                create_pulse(
                    min(flat_segment_end, 0),
                    angular_freq,
                    durations[2],
                    segment_counts[2],
                )
            )

    if end_time < duration:
        # Add trailing step function.
        pwcs.append(
            graph.pwc_signal(duration=duration - end_time, values=np.array([0.0]))
        )

    return graph.time_concatenate_pwc(pwcs, name=name)


@expose(Namespace.PULSES)
def sinusoid_pwc(
    graph: Graph,
    duration: float,
    segment_count: int,
    amplitude: Union[float, complex, Tensor],
    angular_frequency: Union[float, Tensor],
    phase: Union[float, Tensor] = 0.0,
    name: Optional[str] = None,
) -> Pwc:
    r"""
    Creates a `Pwc` representing a sinusoidal oscillation.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    duration : float
        The duration of the signal, :math:`T`.
    segment_count : int
        The number of segments in the PWC.
    amplitude : float or complex or Tensor
        The amplitude of the oscillation, :math:`A`.
        It must either be a scalar or contain a single element.
    angular_frequency : float or Tensor
        The angular frequency of the oscillation, :math:`\omega`.
        It must either be a scalar or contain a single element.
    phase : float or Tensor, optional
        The phase of the oscillation, :math:`\phi`.
        If passed, it must either be a scalar or contain a single element.
        Defaults to 0.
    name : str, optional
        The name of the node.

    Returns
    -------
    Pwc
        The sampled sinusoid.

    See Also
    --------
    :func:`.pulses.cosine_pulse_pwc` : Creates a `Pwc` representing a cosine pulse.

    Notes
    -----
    The sinusoid is defined as

        .. math:: \mathop{\mathrm{Sinusoid}}(t) = A \sin \left( \omega t + \phi \right) \;.

    Examples
    --------

    Define a simple cosine PWC pulse.

    >>> graph.pulses.sinusoid_pwc(
    ...     duration=5.0,
    ...     segment_count=100,
    ...     amplitude=1.0,
    ...     angular_frequency=np.pi,
    ...     phase=np.pi/2.0,
    ...     name="oscillation"
    ... )
    <Pwc: name="oscillation", operation_name="discretize_stf", value_shape=(), batch_shape=()>
    >>> result = qctrl.functions.calculate_graph(graph=graph, output_node_names=["oscillation"])
    >>> result.output["oscillation"]
    [
        {'value': 0.997, 'duration': 0.05},
        {'value': 0.972, 'duration': 0.05},
        ...
        {'value': -0.972, 'duration': 0.05},
        {'value': -0.996, 'duration': 0.05},
    ]

    Define a sinusoid with optimizable parameters.

    >>> graph = qctrl.create_graph()
    >>> amplitude = graph.optimization_variable(
    ...     count=1, lower_bound=0, upper_bound=4e3, name="amplitude"
    ... )
    >>> angular_frequency = graph.optimization_variable(
    ...     count=1, lower_bound=5e6, upper_bound=20e6, name="angular_frequency"
    ... )
    >>> phase = graph.optimization_variable(
    ...     count=1,
    ...     lower_bound=0,
    ...     upper_bound=2*np.pi,
    ...     is_lower_unbounded=True,
    ...     is_upper_unbounded=True,
    ...     name="phase",
    ... )
    >>> graph.pulses.sinusoid_pwc(
    ...     duration=3e-6,
    ...     segment_count=100,
    ...     amplitude=amplitude,
    ...     angular_frequency=angular_frequency,
    ...     phase=phase,
    ...     name="oscillation"
    ... )
    <Pwc: name="oscillation", operation_name="discretize_stf", value_shape=(), batch_shape=()>
    """

    check_argument(
        duration > 0.0, "The duration must be positive.", {"duration": duration}
    )

    amplitude = _validate_optimizable_parameter(graph, amplitude, "amplitude")
    angular_frequency = _validate_optimizable_parameter(
        graph, angular_frequency, "angular frequency"
    )
    phase = _validate_optimizable_parameter(graph, phase, "phase")

    stf = amplitude * graph.sin(angular_frequency * graph.identity_stf() + phase)
    return graph.discretize_stf(stf, duration, segment_count, name=name)


@expose(Namespace.PULSES)
def hann_series_pwc(
    graph: Graph,
    duration: float,
    segment_count: int,
    coefficients: Union[np.ndarray, Tensor],
    name: Optional[str] = None,
) -> Pwc:
    r"""
    Creates a `Pwc` representing a sum of Hann window functions.

    The piecewise-constant function is sampled from Hann functions that start and end at zero.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    duration : float
        The duration of the signal, :math:`T`.
    segment_count : int
        The number of segments in the PWC.
    coefficients : np.ndarray or Tensor
        The coefficients for the different Hann window functions, :math:`c_n`.
        It must be a 1D array or Tensor and it can't contain more than `segment_count` elements.
    name : str, optional
        The name of the node.

    Returns
    -------
    Pwc
        The sampled Hann window functions series.

    See Also
    --------
    :func:`.pulses.cosine_pulse_pwc` : Creates a `Pwc` representing a cosine pulse.

    Notes
    -----

    The series is defined as

        .. math:: \mathop{\mathrm{Hann}}(t)
            = \sum_{n=1}^N c_n \sin^2 \left( \frac{\pi n t}{T} \right) \;,

    where :math:`N` is the number of coefficients.

    Examples
    --------

    Define a simple Hann series.

    >>> graph.pulses.hann_series_pwc(
    ...     duration=5.0,
    ...     segment_count=50,
    ...     coefficients=np.array([0.5, 1, 0.25]),
    ...     name="hann_series",
    ... )
    <Pwc: name="hann_series", operation_name="pwc_signal", value_shape=(), batch_shape=()>
    >>> result = qctrl.functions.calculate_graph(
    ...     graph=graph, output_node_names=["hann_series"]
    ... )
    >>> result.output["hann_series"]
    [
        {'value': 0.0067, 'duration': 0.1},
        {'value': 0.0590, 'duration': 0.1},
        ...
        {'value': 0.0590, 'duration': 0.1},
        {'value': 0.0067, 'duration': 0.1},
    ]

    Define a Hann series with optimizable coefficients.

    >>> coefficients = graph.optimization_variable(
    ...     count=8, lower_bound=-3.5e6, upper_bound=3.5e6, name="coefficients"
    ... )
    >>> graph.pulses.hann_series_pwc(
    ...     duration=2.0e-6,
    ...     segment_count=128,
    ...     coefficients=coefficients,
    ...     name="hann_series",
    ... )
    <Pwc: name="hann_series", operation_name="pwc_signal", value_shape=(), batch_shape=()>
    """

    check_argument(
        duration > 0.0, "The duration must be positive.", {"duration": duration}
    )

    check_argument(
        len(coefficients.shape) == 1,
        "The coefficients must be in a 1D array or Tensor.",
        {"coefficients": coefficients},
        extras={"coefficients.shape": coefficients.shape},
    )

    check_argument(
        coefficients.shape[0] <= segment_count,
        "There can't be more coefficients than segments.",
        {"coefficients": coefficients, "segment_count": segment_count},
        extras={"coefficients.shape": coefficients.shape},
    )

    # Define scaled times π t / T to sample the function.
    times, time_step = np.linspace(
        0, duration, segment_count, endpoint=False, retstep=True
    )
    scaled_times = (times + time_step / 2) * np.pi / duration

    # Calculate function values.
    nss = np.arange(1, coefficients.shape[0] + 1)
    values = graph.sum(coefficients * np.sin(nss * scaled_times[:, None]) ** 2, axis=1)

    return graph.pwc_signal(duration=duration, values=values, name=name)



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

"""qctrltoolkit - a toolkit library for the Q-CTRL Python package."""

__version__ = "1.1.0"
__author__ = "Q-CTRL <support@q-ctrl.com>"

from qctrltoolkit.namespace import (
    TOOLKIT_MAIN_DOC,
    Namespace,
)
from qctrltoolkit.registry import (
    FUNCTIONS,
    NODES,
    TOOLKIT_DOC_CONFIG,
)
from qctrltoolkit.toolkit_utils import (
    TOOLKIT_ATTR,
    forge_toolkit,
)

__all__ = [
    "FUNCTIONS",
    "NODES",
    "TOOLKIT_ATTR",
    "TOOLKIT_DOC_CONFIG",
    "TOOLKIT_MAIN_DOC",
    "Namespace",
    "forge_toolkit",
]



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
System-agnostic convenient nodes.
"""

from typing import Optional

import numpy as np
from qctrlcommons.graph import Graph
from qctrlcommons.node.node_data import Pwc

from qctrltoolkit.namespace import Namespace
from qctrltoolkit.toolkit_utils import expose


@expose(Namespace.UTILS)
def real_optimizable_pwc_signal(
    graph: Graph,
    segment_count: int,
    duration: float,
    maximum: float,
    minimum: float = 0.0,
    name: Optional[str] = None,
) -> Pwc:
    """
    Creates a real optimizable piecewise-constant signal.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    segment_count : int
        The number of piecewise-constant segments in the signal.
    duration : float
        The duration of the signal.
    maximum : float
        The upper bound for the signal values.
    minimum : float, optional
        The lower bound for the signal values. Defaults to 0.
    name : str, optional
        The name of the node.

    Returns
    -------
    Pwc
        The optimizable piecewise-constant signal.

    See Also
    --------
    :func:`.utils.complex_optimizable_pwc_signal` :
        Creates a complex optimizable `Pwc` signal.
    :func:`~qctrl.graphs.Graph.optimization_variable` :
        Creates optimization variables, which can be bounded, semi-bounded, or unbounded.
    """

    values = graph.optimization_variable(
        count=segment_count, lower_bound=minimum, upper_bound=maximum
    )
    return graph.pwc_signal(values=values, duration=duration, name=name)


@expose(Namespace.UTILS)
def complex_optimizable_pwc_signal(
    graph: Graph,
    segment_count: int,
    duration: float,
    maximum: float,
    name: Optional[str] = None,
) -> Pwc:
    """
    Creates a complex optimizable piecewise-constant signal.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    segment_count : int
        The number of segments of the signal.
    duration : float
        The duration of the signal.
    maximum : float
        The upper bound for the modulus of the signal values.
    name : str, optional
        The name of the node.

    Returns
    -------
    Pwc
        The optimizable piecewise-constant signal.

    See Also
    --------
    :func:`~qctrl.graphs.Graph.optimization_variable` :
        Creates optimization variables, which can be bounded, semi-bounded, or unbounded.
    :func:`.utils.real_optimizable_pwc_signal` :
        Creates a real optimizable `Pwc` signal.

    Notes
    -----
    Note that this function sets limits to the modulus of the signal.

    If you want to set (different) limits to the real and imaginary parts instead,
    consider using `graph.util.real_optimizable_signal` to create signals for the
    real and imaginary parts, which you can pass to `graph.complex_value`.
    """

    moduli = graph.optimization_variable(
        count=segment_count,
        lower_bound=0.0,
        upper_bound=maximum,
    )
    phases = graph.optimization_variable(
        count=segment_count,
        lower_bound=-np.pi,
        upper_bound=np.pi,
        is_lower_unbounded=True,
        is_upper_unbounded=True,
    )
    return graph.complex_pwc_signal(
        moduli=moduli, phases=phases, duration=duration, name=name
    )


@expose(Namespace.UTILS)
def filter_and_resample_pwc(
    graph: Graph,
    pwc: Pwc,
    cutoff_frequency: float,
    segment_count: int,
    name: Optional[str] = None,
) -> Pwc:
    """
    Filters a piecewise-constant function with a sinc filter and resamples it again.

    Parameters
    ----------
    graph : Graph
        The graph object where the node will belong.
    pwc : Pwc
        The function to be filtered.
    cutoff_frequency : float
        Upper limit of the range of frequencies that you want to
        preserve in your function.
    segment_count : int
        The number of segments of the resampled filtered function.
    name : str, optional
        The name of the node.

    Returns
    -------
    Pwc
        The filtered and resampled piecewise-constant function.
    """

    return graph.discretize_stf(
        stf=graph.convolve_pwc(
            pwc=pwc, kernel=graph.sinc_convolution_kernel(cutoff_frequency)
        ),
        duration=np.sum(pwc.durations),
        segment_count=segment_count,
        name=name,
    )



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Registry for toolkits.
"""
import inspect
from dataclasses import dataclass
from typing import List

import qctrltoolkit.pulses.functions
import qctrltoolkit.pulses.nodes
import qctrltoolkit.superconducting.functions
import qctrltoolkit.superconducting.nodes
import qctrltoolkit.utils.functions
import qctrltoolkit.utils.nodes
from qctrltoolkit.namespace import Namespace
from qctrltoolkit.toolkit_utils import TOOLKIT_ATTR


def _register(modules):
    """
    Collect exposed toolkits from modules.
    """
    registered = []
    for module in modules:
        for _, member in inspect.getmembers(module):
            if hasattr(member, TOOLKIT_ATTR):
                registered.append(member)
    return registered


NODES = _register(
    [
        qctrltoolkit.superconducting.nodes,
        qctrltoolkit.utils.nodes,
        qctrltoolkit.pulses.nodes,
    ]
)
FUNCTIONS = _register(
    [
        qctrltoolkit.superconducting.functions,
        qctrltoolkit.utils.functions,
        qctrltoolkit.pulses.functions,
    ]
)


def _get_empty_namespaces(exposed_items):
    """
    Returns a list with the names of the namespaces that are not in
    any of the exposed_items' TOOLKIT_ATTRs.
    """
    all_namespaces = set(Namespace)
    exposed_namespaces = set()
    for item in exposed_items:
        exposed_namespaces.update(set(getattr(item, TOOLKIT_ATTR)))

    left_namespaces = all_namespaces.difference(exposed_namespaces)
    return [item.get_name() for item in left_namespaces]


@dataclass
class _DocConfig:
    """
    Class to store information for building docs for the toolkits.

    Parameters
    ----------
    namespaces_without_functions: list[str]
        A list of names for namespaces that don't have functions.
    namespaces_without_nodes : list[str]
        A list of names for namespaces that don't have nodes.
    excluded_class_methods : list[str]
        A list of names for class methods that shouldn't be documented.
    """

    namespaces_without_functions: List[str]
    namespaces_without_nodes: List[str]
    excluded_class_methods: List[str]


TOOLKIT_DOC_CONFIG = _DocConfig(
    namespaces_without_functions=_get_empty_namespaces(FUNCTIONS),
    namespaces_without_nodes=_get_empty_namespaces(NODES),
    excluded_class_methods=["get_pwc"],
)



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Utility functions for exposing and binding toolkits.
"""
import inspect
from functools import partial
from typing import (
    Any,
    Callable,
    List,
    Optional,
    Union,
)

import forge
from numpydoc.docscrape import NumpyDocString
from qctrlcommons.exceptions import QctrlException

from qctrltoolkit.namespace import Namespace

# parameters like qctrl/graph in the toolkit functions are just placeholder for the actual object
# need to remove them from the doc/signature before binding in python
_EXCLUDE_PARAMETERS = {"qctrl", "graph"}

TOOLKIT_ATTR = "_toolkit_attr"


def expose(namespaces: Union[Namespace, List[Namespace]]):
    """
    Define a decorator that inserts a label of the namespaces for later binding to the
    wrapped object.

    The decorator sets a new attribute defined by TOOLKIT_ATTR to the object, which is a tuple
    of strings defining the namespaces. You can apply this decorator to the toolkit function/class,
    and they will be bound to the corresponding namespaces when initializing the Qctrl/Graph object.
    It also sets a new attribute to the object to allow tracking through the python package.

    For example, to expose a function to one namespace ::

        @expose(Namespace.SUPERCONDUCTING)
        def some_func()

    To expose a function to multiple namespaces, as it maybe be reused for
    different physical systems ::

        @expose([Namespace.SUPERCONDUCTING, Namespace.IONS])
        def some_general_func()
    """
    if not isinstance(namespaces, list):
        namespaces = [namespaces]
    return partial(_expose, attr=tuple(namespaces))


def _expose(obj, attr):
    setattr(obj, TOOLKIT_ATTR, attr)
    namespaces = "_".join([f"{i.name.lower()}" for i in attr])
    method_origin = f"qctrl-toolkit.{namespaces}.{getattr(obj, '__name__')}"
    setattr(obj, "method_origin", method_origin)
    return obj


def forge_toolkit(toolkit: Any, obj: Any) -> Any:
    """
    Forge the toolkit object (must be a function or class) for binding.

    Note that when toolkit is a class, we just return it, since it should
    not accept the to-be-bound object, to avoid the cyclic access.

    Parameters
    ----------
    toolkit : Any
        A toolkit function or class to be forged.
    obj: Any
        The object to which the forged toolkit is bound.

    Returns
    -------
    Any
        A forged class/function.

    Raises
    ------
    QctrlException
        If toolkit has an unsupported type.
    """

    if inspect.isfunction(toolkit):
        return _forge_func(toolkit, obj)
    if inspect.isclass(toolkit):
        return toolkit
    raise QctrlException(
        f"Toolkit must either be a function or a class, but got {type(toolkit)}."
    )


def _forge_func(func: Callable, obj: Any) -> Callable:
    """
    Forge toolkit function.
    """

    def forged(*arg, **kwargs):
        if hasattr(obj, "method_origin"):
            obj.method_origin = getattr(func, "method_origin", None)
            result = func(obj, *arg, **kwargs)
            obj.method_origin = None
        else:
            result = func(obj, *arg, **kwargs)

        return result

    _args, _kwargs, _return_type, bind_obj = _get_args_kwargs_return_type(func)

    if not bind_obj:
        return func

    assert func.__doc__ is not None, f"{func} has no doc."
    forged.__doc__ = _clean_doc(func.__doc__)
    forged.__name__ = func.__name__
    sig = forge.sign(*_args, **_kwargs)
    forged = sig(forged)
    forged = forge.returns(_return_type)(forged)

    return forged


def _get_args_kwargs_return_type(func):
    """
    Get the information about the function signature.

    Returns a tuple, where the first element is list of args,
    the second one is a dictionary for kwargs,
    the third one is the type of the returned value,
    and the last one is a boolean to exclude parameters.
    This will be used to forge a new function for binding.
    """
    _args = []
    _kwargs = {}
    bind_obj = False
    _func_inspected = inspect.signature(func)
    for arg_name, _arg in _func_inspected.parameters.items():
        # note that some convenience function might not be fetchable
        if arg_name == "name":
            _kwargs = {"name": forge.kwarg("name", type=Optional[str], default=None)}
        elif arg_name in _EXCLUDE_PARAMETERS:
            bind_obj = True
        else:
            _args.append(
                forge.arg(arg_name, type=_arg.annotation, default=_arg.default)
            )

    return _args, _kwargs, _func_inspected.return_annotation, bind_obj


def _clean_doc(doc: str) -> str:
    """
    Removes unnecessary part from the doc.
    """
    doc_obj = NumpyDocString(doc)

    doc_obj["Parameters"] = [
        item for item in doc_obj["Parameters"] if item.name not in _EXCLUDE_PARAMETERS
    ]

    return doc_obj.__str__()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['qctrltoolkit',
 'qctrltoolkit.pulses',
 'qctrltoolkit.superconducting',
 'qctrltoolkit.utils']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.21.5,<2.0.0',
 'numpydoc>=1.1.0,<2.0.0',
 'python-forge>=18.6.0,<19.0.0',
 'qctrl-commons>=16.3.1,<17.0.0',
 'scipy>=1.7.3,<2.0.0',
 'toml>=0.10.0,<0.11.0']

setup_kwargs = {
    'name': 'qctrl-toolkit',
    'version': '1.1.0',
    'description': 'Q-CTRL Python Toolkit',
    'long_description': '# Q-CTRL Python Toolkit\n\nToolkit of convenience functions and classes for the Q-CTRL Python package.\n',
    'author': 'Q-CTRL',
    'author_email': 'support@q-ctrl.com',
    'maintainer': 'Q-CTRL',
    'maintainer_email': 'support@q-ctrl.com',
    'url': 'https://q-ctrl.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Convenient functions for superconducting systems.
"""

from abc import (
    ABC,
    abstractmethod,
)
from collections import namedtuple
from dataclasses import dataclass
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

import numpy as np
from qctrlcommons.exceptions import QctrlArgumentsValueError
from qctrlcommons.graph import Graph
from qctrlcommons.preconditions import (
    check_argument,
    check_argument_positive_integer,
)

from qctrltoolkit.namespace import Namespace
from qctrltoolkit.toolkit_utils import expose


class OptimizableCoefficient(ABC):
    """
    Abstract class for optimizable Hamiltonian coefficients.
    """

    @abstractmethod
    def get_pwc(self, graph, gate_duration, name):
        """
        Return a Pwc representation of the optimizable coefficient.
        """
        raise NotImplementedError


@expose(Namespace.SUPERCONDUCTING)
@dataclass
class RealOptimizableConstant(OptimizableCoefficient):
    """
    A real-valued optimizable constant coefficient for a Hamiltonian term.
    The main function will try to find the optimal value for this constant.

    Parameters
    ----------
    min : float
        The minimum value that the coefficient can take.
    max : float
        The maximum value that the coefficient can take.
    """

    min: float
    max: float

    def __post_init__(self):
        check_argument(
            self.min < self.max,
            "The maximum must be larger than the minimum.",
            {"min": self.min, "max": self.max},
        )

    def get_pwc(self, graph, gate_duration, name):
        value = graph.optimization_variable(1, self.min, self.max)[0]
        value.name = name
        return graph.constant_pwc(constant=value, duration=gate_duration)


@expose(Namespace.SUPERCONDUCTING)
@dataclass
class ComplexOptimizableConstant(OptimizableCoefficient):
    """
    A complex-valued optimizable constant coefficient for a Hamiltonian term.
    The main function will try to find the optimal value for this constant.

    Parameters
    ----------
    min_modulus : float
        The minimum value that the modulus of the coefficient can take.
    max_modulus : float
        The maximum value that the modulus of the coefficient can take.
    """

    min_modulus: float
    max_modulus: float

    def __post_init__(self):
        check_argument(
            self.min_modulus < self.max_modulus,
            "The maximum must be larger than the minimum.",
            {"min_modulus": self.min_modulus, "max_modulus": self.max_modulus},
        )

    def get_pwc(self, graph, gate_duration, name):
        mod = graph.optimization_variable(1, self.min_modulus, self.max_modulus)[0]
        phase = graph.optimization_variable(1, 0, 2 * np.pi, True, True)[0]
        value = graph.multiply(mod, graph.exp(1j * phase), name=name)
        return graph.constant_pwc(constant=value, duration=gate_duration)


@expose(Namespace.SUPERCONDUCTING)
@dataclass
class RealOptimizableSignal(OptimizableCoefficient):
    """
    A real-valued optimizable time-dependent piecewise-constant coefficient for
    a Hamiltonian term. The main function will try to find the optimal value for
    this signal at each segment.

    Parameters
    ----------
    count : int
        The number of segments in the piecewise-constant signal.
    min : float
        The minimum value that the signal can take at each segment.
    max : float
        The maximum value that the signal can take at each segment.
    """

    count: int
    min: float
    max: float

    def __post_init__(self):
        check_argument(
            self.count > 0, "There must be at least one segment.", {"count": self.count}
        )
        check_argument(
            self.min < self.max,
            "The maximum must be larger than the minimum.",
            {"min": self.min, "max": self.max},
        )

    def get_pwc(self, graph, gate_duration, name):
        values = graph.optimization_variable(self.count, self.min, self.max)
        return graph.pwc_signal(values=values, duration=gate_duration, name=name)


@expose(Namespace.SUPERCONDUCTING)
@dataclass
class ComplexOptimizableSignal(OptimizableCoefficient):
    """
    A complex-valued optimizable time-dependent piecewise-constant coefficient
    for a Hamiltonian term. The main function will try to find the optimal value
    for this signal at each segment.

    Parameters
    ----------
    count : int
        The number of segments in the piecewise-constant signal.
    min_modulus : float
        The minimum value that the modulus of the signal can take at each segment.
    max_modulus : float
        The maximum value that the modulus of the signal can take at each segment.
    """

    count: int
    min_modulus: float
    max_modulus: float

    def __post_init__(self):
        check_argument(
            self.count > 0, "There must be at least one segment.", {"count": self.count}
        )
        check_argument(
            self.min_modulus < self.max_modulus,
            "The maximum must be larger than the minimum.",
            {"min_modulus": self.min_modulus, "max_modulus": self.max_modulus},
        )

    def get_pwc(self, graph, gate_duration, name):
        mods = graph.optimization_variable(
            self.count, self.min_modulus, self.max_modulus
        )
        phases = graph.optimization_variable(self.count, 0, 2 * np.pi, True, True)
        return graph.pwc_signal(
            values=mods * graph.exp(1j * phases), duration=gate_duration, name=name
        )


RealCoefficient = Union[
    int, float, np.ndarray, RealOptimizableSignal, RealOptimizableConstant
]
Coefficient = Union[
    RealCoefficient, complex, ComplexOptimizableSignal, ComplexOptimizableConstant
]


def _check_argument_real_coefficient(argument, name):

    if isinstance(argument, (ComplexOptimizableSignal, ComplexOptimizableConstant)):
        check = False
    else:
        check = np.isrealobj(argument)

    check_argument(check, f"The {name} must be a real coefficient.", {name: argument})


@expose(Namespace.SUPERCONDUCTING)
@dataclass
class Transmon:
    """
    Class that stores all the physical system data for a transmon.

    Parameters
    ----------
    dimension : int
        Number of dimensions of the Hilbert space of the transmon.
        Must be at least 2.
    frequency : real or np.ndarray or RealOptimizableSignal or RealOptimizableConstant, optional
        The frequency of the transmon, :math:`\\omega_t`.
        If not provided, it defaults to no frequency term.
    anharmonicity : real or np.ndarray or RealOptimizableSignal or RealOptimizableConstant, optional
        The nonlinearity of the transmon, :math:`\\alpha`.
        If not provided, it defaults to no anharmonicity term.
    drive : real or complex or np.ndarray or RealOptimizableSignal or \
            RealOptimizableConstant or ComplexOptimizableSignal or \
            ComplexOptimizableConstant, optional
        The complex drive of the transmon, :math:`\\gamma_t`.
        If not provided, it defaults to no drive term.
    name : str, optional
        The identifier of the transmon that is used to link interaction terms to this transmon.
        Defaults to "transmon".

    Notes
    -----
    The Hamiltonian for the transmon is defined as

    .. math::
        H_\\mathrm{transmon} =
            \\omega_t b^\\dagger b
            + \\frac{\\alpha}{2} (b^\\dagger)^2 b^2
            + \\frac{1}{2} \\left(\\gamma_t b^\\dagger + H.c. \\right) \\;,

    where :math:`H.c.` indicates the Hermitian conjugate.
    All coefficients in the Hamiltonian are optional,
    and you should only pass those relevant to your system.
    """

    dimension: int
    frequency: Optional[RealCoefficient] = None
    anharmonicity: Optional[RealCoefficient] = None
    drive: Optional[Coefficient] = None
    name: str = "transmon"

    def __post_init__(self):
        check_argument_positive_integer(self.dimension, "dimension")
        check_argument(
            self.dimension >= 2,
            "The dimension must be at least 2.",
            {"dimension": self.dimension},
        )
        _check_argument_real_coefficient(self.frequency, "frequency")
        _check_argument_real_coefficient(self.anharmonicity, "anharmonicity")


@expose(Namespace.SUPERCONDUCTING)
@dataclass
class Cavity:
    """
    Class that stores all the physical system data for a cavity.

    Parameters
    ----------
    dimension : int
        Number of dimensions of the Hilbert space of the cavity.
        Must be at least 2.
    frequency : real or np.ndarray or RealOptimizableSignal or RealOptimizableConstant, optional
        The frequency of the cavity, :math:`\\omega_c`.
        If not provided, it defaults to no frequency term.
    kerr_coefficient : real or np.ndarray or RealOptimizableSignal or \
            RealOptimizableConstant, optional
        The nonlinearity of the cavity, :math:`K`.
        If not provided, it defaults to no nonlinear term.
    drive : real or complex or np.ndarray or RealOptimizableSignal or \
            RealOptimizableConstant or ComplexOptimizableSignal or \
            ComplexOptimizableConstant, optional
        The complex drive of the cavity, :math:`\\gamma_c`.
        If not provided, it defaults to no drive term.
    name : str, optional
        The identifier of the cavity that is used to link interaction terms to this cavity.
        Defaults to "cavity".

    Notes
    -----
    The Hamiltonian for the cavity is defined as

    .. math::
        H_\\mathrm{cavity} =
            \\omega_c a^\\dagger a
            + \\frac{K}{2} (a^\\dagger)^2 a^2
            + \\frac{1}{2} \\left(\\gamma_c a^\\dagger + H.c. \\right) \\;,

    where :math:`H.c.` indicates the Hermitian conjugate.
    All coefficients in the Hamiltonian are optional,
    and you should only pass those relevant to your system.
    """

    dimension: int
    frequency: Optional[RealCoefficient] = None
    kerr_coefficient: Optional[RealCoefficient] = None
    drive: Optional[Coefficient] = None
    name: str = "cavity"

    def __post_init__(self):
        check_argument_positive_integer(self.dimension, "dimension")
        check_argument(
            self.dimension >= 2,
            "The dimension must be at least 2.",
            {"dimension": self.dimension},
        )
        _check_argument_real_coefficient(self.frequency, "frequency")
        _check_argument_real_coefficient(self.kerr_coefficient, "Kerr coefficient")


@expose(Namespace.SUPERCONDUCTING)
@dataclass
class TransmonTransmonInteraction:
    """
    Class that stores all the physical system data for the interaction
    between two transmons.

    Parameters
    ----------
    transmon_names : tuple[str, str]
        The two names identifying the transmons in the interaction.
    effective_coupling : real or complex or np.ndarray or RealOptimizableSignal or \
            RealOptimizableConstant or ComplexOptimizableSignal or \
            ComplexOptimizableConstant, optional
        The effective coupling between the two transmons, :math:`g`.
        If not provided, it defaults to no effective coupling term.

    Notes
    -----
    The Hamiltonian for the interaction is defined as

    .. math::
        H_\\mathrm{transmon-transmon} = g b_1 b_2^\\dagger + H.c. \\;.
    """

    transmon_names: Tuple[str, str]
    effective_coupling: Optional[Coefficient]

    def __post_init__(self):
        check_argument(
            self.transmon_names[0] != self.transmon_names[1],
            "The names of the two transmons must be different.",
            {"transmon_names": self.transmon_names},
        )


@expose(Namespace.SUPERCONDUCTING)
@dataclass
class TransmonCavityInteraction:
    """
    Class that stores all the physical system data for the interaction
    between a transmon and a cavity.

    Parameters
    ----------
    dispersive_shift : real or np.ndarray or RealOptimizableSignal or \
            RealOptimizableConstant, optional
        The dispersive shift between the transmon and the cavity, :math:`\\chi`.
        You must provide either a dispersive shift or a Rabi coupling.
    rabi_coupling : real or complex or np.ndarray or RealOptimizableSignal or \
            RealOptimizableConstant or ComplexOptimizableSignal or \
            ComplexOptimizableConstant, optional
        The strength of the Rabi coupling between the transmon and the cavity, :math:`\\Omega`.
        You must provide either a dispersive shift or a Rabi coupling.
    transmon_name : str, optional
        The name identifying the transmon in the interaction.
        Defaults to "transmon".
    cavity_name : str, optional
        The name identifying the cavity in the interaction.
        Defaults to "cavity".

    Notes
    -----
    The Hamiltonian for the interaction is defined as

    .. math:: H_\\mathrm{transmon-cavity} = \\chi a^\\dagger a b^\\dagger b \\;,

    or as

    .. math:: H_\\mathrm{transmon-cavity} = \\Omega a b^\\dagger + H.c. \\;,

    where :math:`H.c.` indicates the Hermitian conjugate.
    """

    dispersive_shift: Optional[RealCoefficient] = None
    rabi_coupling: Optional[Coefficient] = None
    transmon_name: Optional[str] = "transmon"
    cavity_name: Optional[str] = "cavity"

    def __post_init__(self):
        _check_argument_real_coefficient(self.dispersive_shift, "dispersive shift")
        check_argument(
            (self.dispersive_shift is None) ^ (self.rabi_coupling is None),  # xor
            "You must provide either a dispersive shift or a Rabi coupling.",
            {
                "dispersive_shift": self.dispersive_shift,
                "rabi_coupling": self.rabi_coupling,
            },
        )


@expose(Namespace.SUPERCONDUCTING)
@dataclass
class CavityCavityInteraction:
    r"""
    Class that stores all the physical system data for the interaction
    between two cavities.

    Parameters
    ----------
    cavity_names : tuple[str, str]
        The two names identifying the cavities in the interaction.
    cross_kerr_coefficient : real or np.ndarray or RealOptimizableSignal or RealOptimizableConstant
        The cross-Kerr coefficient between the two cavities, :math:`K_{12}`.
        If not provided, it defaults to no cross-Kerr term.

    Notes
    -----
    The Hamiltonian for the interaction is defined as

    .. math::
        H_\mathrm{cavity-cavity} = K_{12} a_1^\dagger a_1 a_2^\dagger a_2 \;.
    """

    cavity_names: Tuple[str, str]
    cross_kerr_coefficient: RealCoefficient

    def __post_init__(self):
        check_argument(
            self.cavity_names[0] != self.cavity_names[1],
            "The names of the two cavities must be different.",
            {"cavity_names": self.cavity_names},
        )
        _check_argument_real_coefficient(
            self.cross_kerr_coefficient, "cross-Kerr coefficient"
        )


Interaction = Union[
    TransmonTransmonInteraction, TransmonCavityInteraction, CavityCavityInteraction
]


def _check_real_coefficient(obj):
    if isinstance(obj, (RealOptimizableSignal, RealOptimizableConstant)):
        return True

    if np.isscalar(obj) or isinstance(obj, np.ndarray):
        return np.isrealobj(obj)

    return False


_QHOOps = namedtuple("_QHOOps", ["a", "adag", "n"])


def _create_qho_operators(
    graph, transmons: List[Transmon], cavities: List[Cavity]
) -> Tuple[Dict[str, _QHOOps], Dict[str, _QHOOps]]:
    """
    Creates the creation, annihilation, and number operators for transmons and cavities.

    Returns a dictionary whose keys are the subsystem names and whose values are _QHOOps
    namedtuples (with field names a, adag, and n) with the operators of each subsystem.
    """

    # Dimension of the identity in the Kronecker product before and after the operator.
    pre_dim = 1
    post_dim = np.prod([system.dimension for system in transmons + cavities])

    tri_kron = lambda op1, op2, op3: graph.kron(graph.kron(op1, op2), op3)

    # Define operators for each subsystem.
    operators = {}
    for system in transmons + cavities:
        dim = system.dimension
        post_dim //= dim
        pre_eye = graph.tensor(np.eye(pre_dim))
        post_eye = graph.tensor(np.eye(post_dim))
        operators[system.name] = _QHOOps(
            tri_kron(pre_eye, graph.annihilation_operator(dim), post_eye),
            tri_kron(pre_eye, graph.creation_operator(dim), post_eye),
            tri_kron(pre_eye, graph.number_operator(dim), post_eye),
        )
        pre_dim *= dim

    return operators


def _validate_physical_system_inputs(
    transmons: List[Transmon],
    cavities: List[Cavity],
    interactions: List[Interaction],
) -> Tuple[
    List[TransmonTransmonInteraction],
    List[TransmonCavityInteraction],
    List[CavityCavityInteraction],
]:
    """
    Validates the type of the subsystems and interactions.
    Validates the subsystem names in the interactions.

    Returns three lists containing
    (1) all the transmon-transmon interactions,
    (2) all the transmon-cavity interactions, and
    (3) all the cavity-cavity interactions.
    """

    check_argument(
        len(transmons + cavities) > 0,
        "At least one transmon or cavity must be provided.",
        {"transmons": transmons, "cavities": cavities},
    )

    check_argument(
        all(isinstance(transmon, Transmon) for transmon in transmons),
        "Each element in transmons must be a Transmon object.",
        {"transmons": transmons},
    )
    transmon_names = [transmon.name for transmon in transmons]
    check_argument(
        len(transmon_names) == len(set(transmon_names)),
        "Transmon names must be unique.",
        {"transmons": transmons},
        extras={"[transmon.name for transmon in transmons]": transmon_names},
    )

    check_argument(
        all(isinstance(cavity, Cavity) for cavity in cavities),
        "Each element in cavities must be a Cavity object.",
        {"cavities": cavities},
    )
    cavity_names = [cavity.name for cavity in cavities]
    check_argument(
        len(cavity_names) == len(set(cavity_names)),
        "Cavity names must be unique.",
        {"cavities": cavities},
        extras={"[cavity.name for cavity in cavities]": cavity_names},
    )

    check_argument(
        len(transmon_names + cavity_names) == len(set(transmon_names + cavity_names)),
        "Transmon and cavity names must be unique.",
        {"transmons": transmons, "cavities": cavities},
        extras={
            "[transmon.name for transmon in transmons]": transmon_names,
            "[cavity.name for cavity in cavities]": cavity_names,
        },
    )

    transmon_transmon_interactions = [
        intx for intx in interactions if isinstance(intx, TransmonTransmonInteraction)
    ]
    transmon_cavity_interactions = [
        intx for intx in interactions if isinstance(intx, TransmonCavityInteraction)
    ]
    cavity_cavity_interactions = [
        intx for intx in interactions if isinstance(intx, CavityCavityInteraction)
    ]
    check_argument(
        len(
            transmon_transmon_interactions
            + transmon_cavity_interactions
            + cavity_cavity_interactions
        )
        == len(interactions),
        "Each element in interactions must be a TransmonTransmonInteraction, "
        "a TransmonCavityInteraction, or a CavityCavityInteraction object.",
        {"interactions": interactions},
    )

    tt_interaction_name_pairs = set(
        frozenset(tt_interaction.transmon_names)
        for tt_interaction in transmon_transmon_interactions
    )
    check_argument(
        len(tt_interaction_name_pairs) == len(transmon_transmon_interactions),
        "There are duplicate transmon-transmon interaction terms.",
        {"interactions": interactions},
    )

    for tt_interaction in transmon_transmon_interactions:
        name_1, name_2 = tt_interaction.transmon_names
        check_argument(
            name_1 in transmon_names and name_2 in transmon_names,
            "Names in transmon-transmon interaction terms must refer to transmons in the system.",
            {"transmons": transmons, "interactions": interactions},
            extras={"[transmon.name for transmon in transmons]": transmon_names},
        )

    tc_interaction_name_pairs = set(
        (tc_interaction.transmon_name, tc_interaction.cavity_name)
        for tc_interaction in transmon_cavity_interactions
    )
    check_argument(
        len(tc_interaction_name_pairs) == len(transmon_cavity_interactions),
        "There are duplicate transmon-cavity interaction terms.",
        {"interactions": interactions},
    )

    for tc_interaction in transmon_cavity_interactions:
        check_argument(
            tc_interaction.transmon_name in transmon_names,
            "Transmon names in transmon-cavity interaction terms "
            "must refer to transmons in the system.",
            {"transmons": transmons, "interactions": interactions},
            extras={"[transmon.name for transmon in transmons]": transmon_names},
        )
        check_argument(
            tc_interaction.cavity_name in cavity_names,
            "Cavity names in transmon-cavity interaction terms "
            "must refer to cavities in the system.",
            {"cavities": cavities, "interactions": interactions},
            extras={"[cavity.name for cavity in cavities]": cavity_names},
        )

    cc_interaction_name_pairs = set(
        frozenset(cc_interaction.cavity_names)
        for cc_interaction in cavity_cavity_interactions
    )
    check_argument(
        len(cc_interaction_name_pairs) == len(cavity_cavity_interactions),
        "There are duplicate cavity-cavity interaction terms.",
        {"interactions": interactions},
    )

    for cc_interaction in cavity_cavity_interactions:
        name_1, name_2 = cc_interaction.cavity_names
        check_argument(
            name_1 in cavity_names and name_2 in cavity_names,
            "Names in cavity-cavity interaction terms must refer to cavities in the system.",
            {"cavities": cavities, "interactions": interactions},
            extras={"[cavity.name for cavity in cavities]": cavity_names},
        )

    return (
        transmon_transmon_interactions,
        transmon_cavity_interactions,
        cavity_cavity_interactions,
    )


def _create_superconducting_hamiltonian(
    graph: Graph,
    transmons: List[Transmon],
    cavities: List[Cavity],
    interactions: List[Interaction],
    gate_duration: float,
    cutoff_frequency: float,
    sample_count: int,
):
    r"""
    Creates the Hamiltonian of a system composed of a single transmon and one or more cavities.
    Returns the Hamiltonian as a Pwc node and a list with the names of the optimizable nodes
    that have been added to the graph.

    Parameters
    ----------
    graph : Graph
        The graph where the Hamiltonian will be added.
    transmons : list[Transmon]
        List of objects containing the physical information about the transmons.
    cavities : list[Cavity]
        List of objects containing the physical information about the cavities.
    interactions : list[Interaction]
        List of objects containing the physical information about the interactions in the system.
    gate_duration : float
        The duration of the gate.
    cutoff_frequency : float
        The cutoff frequency of a linear sinc filter to be applied to the piecewise-constant
        signals you provide for the coefficients. If None, the signals are not filtered.
    sample_count : int
        The number of segments into which the PWC terms are discretized.

    Returns
    -------
    Pwc
        A node representing the system's Hamiltonian.
    list
        The names of the graph nodes representing optimizable coefficients.
        If some of these are PWC functions and cutoff_frequency is not None,
        then the names of the filtered PWC nodes are also included.
    """

    def convert_to_pwc(coefficient: Coefficient, real_valued: bool, name: str):
        """
        Returns the Pwc representation of a coefficient.
        """

        def filter_signal(signal):
            return graph.discretize_stf(
                stf=graph.convolve_pwc(pwc=signal, kernel=kernel),
                duration=gate_duration,
                segment_count=sample_count,
                name=f"{name}_filtered",
            )

        if real_valued:
            check_argument(
                _check_real_coefficient(coefficient),
                f"{name} can't be complex.",
                {name: coefficient},
            )

        # Convert scalar value into constant Pwc.
        if np.isscalar(coefficient):
            return graph.constant_pwc(
                constant=graph.tensor(coefficient, name=name), duration=gate_duration
            )

        # Convert array into Pwc.
        if isinstance(coefficient, np.ndarray):
            signal = graph.pwc_signal(coefficient, gate_duration, name=name)

            if kernel is None:
                return signal

            return filter_signal(signal)

        # Convert Real/ComplexOptimizableConstant into optimizable constant Pwc.
        if isinstance(
            coefficient, (RealOptimizableConstant, ComplexOptimizableConstant)
        ):
            optimizable_node_names.append(name)
            return coefficient.get_pwc(graph, gate_duration, name)

        # Convert Real/ComplexOptimizableSignal into optimizable Pwc.
        if isinstance(coefficient, (RealOptimizableSignal, ComplexOptimizableSignal)):
            optimizable_node_names.append(name)
            signal = coefficient.get_pwc(graph, gate_duration, name)
            if kernel is None:
                return signal

            optimizable_node_names.append(f"{name}_filtered")
            return filter_signal(signal)

        raise QctrlArgumentsValueError(
            f"{name} has an invalid type.", {name: coefficient}
        )

    (
        transmon_transmon_interactions,
        transmon_cavity_interactions,
        cavity_cavity_interactions,
    ) = _validate_physical_system_inputs(transmons, cavities, interactions)

    # Define annihilation and creation operators for the transmon and the cavity.
    operators = _create_qho_operators(graph, transmons, cavities)

    # Create nested dictionary structure containing information for the different Hamiltonian terms.
    hamiltonian_info = {}

    # Add transmon terms.
    for transmon in transmons:
        transmon_ops = operators[transmon.name]
        hamiltonian_info[f"{transmon.name}.frequency"] = {
            "coefficient": transmon.frequency,
            "operator": transmon_ops.n,
            "is_hermitian": True,
        }
        hamiltonian_info[f"{transmon.name}.anharmonicity"] = {
            "coefficient": transmon.anharmonicity,
            "operator": 0.5 * (transmon_ops.n @ transmon_ops.n - transmon_ops.n),
            "is_hermitian": True,
        }
        hamiltonian_info[f"{transmon.name}.drive"] = {
            "coefficient": transmon.drive,
            "operator": transmon_ops.adag,
            "is_hermitian": False,
        }

    # Add cavity terms.
    for cavity in cavities:
        cavity_ops = operators[cavity.name]
        hamiltonian_info[f"{cavity.name}.frequency"] = {
            "coefficient": cavity.frequency,
            "operator": cavity_ops.n,
            "is_hermitian": True,
        }
        hamiltonian_info[f"{cavity.name}.kerr_coefficient"] = {
            "coefficient": cavity.kerr_coefficient,
            "operator": 0.5 * (cavity_ops.n @ cavity_ops.n - cavity_ops.n),
            "is_hermitian": True,
        }
        hamiltonian_info[f"{cavity.name}.drive"] = {
            "coefficient": cavity.drive,
            "operator": cavity_ops.adag,
            "is_hermitian": False,
        }

    # Add transmon-transmon interaction terms.
    for tt_interaction in transmon_transmon_interactions:
        name_1, name_2 = tt_interaction.transmon_names
        transmon_1_ops = operators[name_1]
        transmon_2_ops = operators[name_2]
        key = f"{name_1}_{name_2}_interaction"
        hamiltonian_info[f"{key}.effective_coupling"] = {
            "coefficient": tt_interaction.effective_coupling,
            "operator": 2 * transmon_1_ops.a @ transmon_2_ops.adag,
            "is_hermitian": False,
        }

    # Add transmon-cavity interaction terms.
    for tc_interaction in transmon_cavity_interactions:
        transmon_ops = operators[tc_interaction.transmon_name]
        cavity_ops = operators[tc_interaction.cavity_name]
        key = f"{tc_interaction.transmon_name}_{tc_interaction.cavity_name}_interaction"
        hamiltonian_info[f"{key}.dispersive_shift"] = {
            "coefficient": tc_interaction.dispersive_shift,
            "operator": transmon_ops.n @ cavity_ops.n,
            "is_hermitian": True,
        }
        hamiltonian_info[f"{key}.rabi_coupling"] = {
            "coefficient": tc_interaction.rabi_coupling,
            "operator": 2 * transmon_ops.adag @ cavity_ops.a,
            "is_hermitian": False,
        }

    # Add cavity-cavity interaction terms.
    for cc_interaction in cavity_cavity_interactions:
        name_1, name_2 = cc_interaction.cavity_names
        cavity_1_ops = operators[name_1]
        cavity_2_ops = operators[name_2]
        key = f"{name_1}_{name_2}_interaction"
        hamiltonian_info[f"{key}.cross_kerr_coefficient"] = {
            "coefficient": cc_interaction.cross_kerr_coefficient,
            "operator": cavity_1_ops.n @ cavity_2_ops.n,
            "is_hermitian": True,
        }

    check_argument(
        any(np.any(info["coefficient"]) for info in hamiltonian_info.values()),
        "The system must contain at least one Hamiltonian coefficient.",
        {"transmons": transmons, "cavities": cavities, "interactions": interactions},
    )

    # Create kernel to filter signals (used in convert_to_pwc).
    if cutoff_frequency is not None:
        kernel = graph.sinc_convolution_kernel(cutoff_frequency)
    else:
        kernel = None

    # Build the Hamiltonian from the different terms.
    hamiltonian_terms = []
    optimizable_node_names = []  # filled up by convert_to_pwc

    for name, info in hamiltonian_info.items():
        if info["coefficient"] is not None:
            coefficient = convert_to_pwc(
                coefficient=info["coefficient"],
                real_valued=info["is_hermitian"],
                name=name,
            )
            if info["is_hermitian"]:
                hamiltonian_terms.append(coefficient * info["operator"])
            else:
                hamiltonian_terms.append(0.5 * coefficient * info["operator"])
                hamiltonian_terms.append(
                    0.5 * graph.conjugate(coefficient) * graph.adjoint(info["operator"])
                )

    return graph.pwc_sum(hamiltonian_terms), optimizable_node_names


def _extract_output(item):
    """
    Converts a non-batched item from a qctrl.types.graph/optimization.Result.output
    dictionary into a NumPy array.

    If the item is a dictionary (corresponds to a tensor), then its "value" is returned.

    If the item is a list of dictionaries (corresponds to a Pwc), then the "value"
    elements of its inner dictionaries are returned in a NumPy array.
    """
    if isinstance(item, dict):
        return item["value"]
    if isinstance(item, list):
        assert all(
            isinstance(elem, dict) for elem in item
        ), "Batches are not supported."
        return np.array([seg["value"] for seg in item])
    assert False, "Unknown output format."


@expose(Namespace.SUPERCONDUCTING)
def simulate(
    qctrl: Any,
    transmons: List[Transmon],
    cavities: List[Cavity],
    interactions: List[Interaction],
    gate_duration: float,
    sample_count: int = 128,
    cutoff_frequency: Optional[float] = None,
    initial_state: Optional[np.ndarray] = None,
):
    """
    Simulates a system composed of transmons and cavities.

    Parameters
    ----------
    qctrl : qctrl.Qctrl
        Boulder Opal session object.
    transmons : list[Transmon]
        List of objects containing the physical information about the transmons.
        It must not contain any optimizable coefficients.
        It can be an empty list, but at least one transmon or cavity must be provided.
    cavities : list[Cavity]
        List of objects containing the physical information about the cavities.
        They must not contain any optimizable coefficients.
        It can be an empty list, but at least one transmon or cavity must be provided.
    interactions : list[TransmonTransmonInteraction or TransmonCavityInteraction or \
            CavityCavityInteraction]
        List of objects containing the physical information about the interactions in the system.
        They must not contain any optimizable coefficients.
        It can be an empty list.
    gate_duration : float
        The duration of the gate to be simulated, :math:`t_\\mathrm{gate}`.
        It must be greater than zero.
    sample_count : int, optional
        The number of times between 0 and `gate_duration` (included)
        at which the evolution is sampled.
        Defaults to 128.
    cutoff_frequency : float, optional
        The cutoff frequency of a linear sinc filter to be applied to the piecewise-constant
        signals you provide for the coefficients. If not provided, the signals are not filtered.
        If the signals are filtered, a larger sample count leads to a more accurate numerical
        integration. (If the signals are not filtered, the sample count has no effect on the
        numerical precision of the integration.)
    initial_state : np.ndarray, optional
        The initial state of the system, :math:`|\\Psi_\\mathrm{initial}\\rangle`, as a 1D array of
        length ``D = np.prod([system.dimension for system in transmons + cavities])``.
        If not provided, the function only returns the system's unitary time-evolution operators.

    Returns
    -------
    dict
        A dictionary containing information about the time evolution of the system, with keys

            ``sample_times``
                The times at which the system's evolution is sampled,
                as an array of shape ``(T,)``.
            ``unitaries``
                The system's unitary time-evolution operators at each sample time,
                as an array of shape ``(T, D, D)``.
            ``state_evolution``
                The time evolution of the initial state at each sample time,
                as an array of shape ``(T, D)``.
                This is only returned if you provide an initial state.

    Notes
    -----
    The Hamiltonian of the system is of the form

    .. math::
        H = \\sum_i H_{\\mathrm{transmon}_i}
            + \\sum_i H_{\\mathrm{cavity}_i}
            + \\sum_{i,j} H_{\\mathrm{transmon}_i-\\mathrm{transmon}_j}
            + \\sum_{i,j} H_{\\mathrm{transmon}_i-\\mathrm{cavity}_j}
            + \\sum_{i,j} H_{\\mathrm{cavity}_i-\\mathrm{cavity}_j}

    where i and j mark the i-th and j-th transmon or cavity.
    For their definition of each Hamiltonian term, see its respective class.

    The Hilbert space of the system is defined as the outer product of all the
    transmon Hilbert spaces (in the order they're provided in `transmons`) with
    the cavity Hilbert spaces (in the order they're provided in `cavities`), that is:

    .. math::
        \\mathcal{H} =
            \\mathcal{H}_{\\mathrm{transmon}_1} \\otimes \\mathcal{H}_{\\mathrm{transmon}_2}
            \\otimes \\ldots
            \\otimes \\mathcal{H}_{\\mathrm{cavity}_1} \\otimes \\mathcal{H}_{\\mathrm{cavity}_2}
            \\otimes \\ldots

    The system dimension `D` is then the product of all transmon and cavity dimensions.
    """
    system_dimension = np.prod(
        [system.dimension for system in transmons + cavities], dtype=int
    )

    if initial_state is not None:
        check_argument(
            initial_state.shape == (system_dimension,),
            "Initial state must be a 1D array of length "
            "np.prod([system.dimension for system in transmons + cavities]).",
            {
                "initial_state": initial_state,
                "transmons": transmons,
                "cavities": cavities,
            },
            extras={
                "initial_state.shape": initial_state.shape,
                "np.prod([system.dimension for system in transmons + cavities])": system_dimension,
            },
        )
    check_argument(
        gate_duration > 0,
        "The gate duration must be greater than zero.",
        {"gate_duration": gate_duration},
    )

    # Create graph object.
    graph = qctrl.create_graph()

    # Create PWC Hamiltonian.
    hamiltonian, optimizable_node_names = _create_superconducting_hamiltonian(
        graph=graph,
        transmons=transmons,
        cavities=cavities,
        interactions=interactions,
        gate_duration=gate_duration,
        cutoff_frequency=cutoff_frequency,
        sample_count=sample_count,
    )

    # Check whether there are any optimizable coefficients.
    check_argument(
        len(optimizable_node_names) == 0,
        "None of the Hamiltonian terms can be optimizable.",
        {"transmons": transmons, "cavities": cavities, "interactions": interactions},
    )

    # Calculate the evolution.
    sample_times = np.linspace(
        gate_duration / sample_count, gate_duration, sample_count
    )
    unitaries = graph.time_evolution_operators_pwc(
        hamiltonian=hamiltonian, sample_times=sample_times, name="unitaries"
    )
    output_node_names = ["unitaries"]

    if initial_state is not None:
        states = unitaries @ initial_state[:, None]
        states = states[..., 0]
        states.name = "state_evolution"
        output_node_names.append("state_evolution")

    simulation_result = qctrl.functions.calculate_graph(
        graph=graph, output_node_names=output_node_names
    )

    # Retrieve results and build output dictionary.
    result_dict = {"sample_times": sample_times}

    for key in output_node_names:
        result_dict[key] = _extract_output(simulation_result.output[key])

    return result_dict


@expose(Namespace.SUPERCONDUCTING)
def optimize(
    qctrl: Any,
    transmons: List[Transmon],
    cavities: List[Cavity],
    interactions: List[Interaction],
    gate_duration: float,
    initial_state: Optional[np.ndarray] = None,
    target_state: Optional[np.ndarray] = None,
    target_operation: Optional[np.ndarray] = None,
    sample_count: int = 128,
    cutoff_frequency: Optional[float] = None,
    target_cost: Optional[float] = None,
    optimization_count: int = 5,
) -> Dict[str, Any]:
    """
    Finds optimal pulses or parameters for a system composed of transmons and cavities,
    in order to achieve a target state or implement a target operation.

    At least one of the terms in the transmon, cavities, or interaction Hamiltonians
    must be optimizable.

    To optimize a state transfer, you need to provide an initial and a target state.
    To optimize a target gate/unitary, you need to provide a target operation.

    Parameters
    ----------
    qctrl : qctrl.Qctrl
        Boulder Opal session object.
    transmons : list[Transmon]
        List of objects containing the physical information about the transmons.
        It can be an empty list, but at least one transmon or cavity must be provided.
    cavities : list[Cavity]
        List of objects containing the physical information about the cavities.
        It can be an empty list, but at least one transmon or cavity must be provided.
    interactions : list[TransmonTransmonInteraction or TransmonCavityInteraction or \
            CavityCavityInteraction]
        List of objects containing the physical information about the interactions in the system.
        It can be an empty list.
    gate_duration : float
        The duration of the gate to be optimized, :math:`t_\\mathrm{gate}`.
        It must be greater than zero.
    initial_state : np.ndarray, optional
        The initial state of the system, :math:`|\\Psi_\\mathrm{initial}\\rangle`, as a 1D array of
        length ``D = np.prod([system.dimension for system in transmon + cavities])``.
        If provided, the function also returns its time evolution.
        This is a required parameter if you pass a `target_state`.
    target_state : np.ndarray, optional
        The target state of the optimization, :math:`|\\Psi_\\mathrm{target}\\rangle`,
        as a 1D array of length `D`.
        You must provide exactly one of `target_state` or `target_operation`.
    target_operation : np.ndarray, optional
        The target operation of the optimization, :math:`U_\\mathrm{target}`,
        as a 2D array of shape ``(D, D)``.
        You must provide exactly one of `target_state` or `target_operation`.
    sample_count : int, optional
        The number of times between 0 and `gate_duration` (included)
        at which the evolution is sampled.
        Defaults to 128.
    cutoff_frequency : float, optional
        The cutoff frequency of a linear sinc filter to be applied to the piecewise-constant
        signals you provide for the coefficients. If not provided, the signals are not filtered.
        If the signals are filtered, a larger sample count leads to a more accurate numerical
        integration. (If the signals are not filtered, the sample count has no effect on the
        numerical precision of the integration.)
    target_cost : float, optional
        A target value of the cost that you can set as an early stop condition for the optimizer.
        If not provided, the optimizer runs until it converges.
    optimization_count : int, optional
        The number of independent randomly seeded optimizations to perform. The result
        from the best optimization (the one with the lowest cost) is returned.
        Defaults to 5.

    Returns
    -------
    dict
        A dictionary containing the optimized coefficients and information about the time evolution
        of the system, with keys

            ``optimized_values``
                Dictionary containing the requested optimized Hamiltonian coefficients, under
                keys such as ``[transmon_1_name].drive``, ``[cavity_2_name].frequency``,
                ``[transmon_2_name]_[cavity_1_name]_interaction.dispersive_shift``, and
                ``[cavity_1_name]_[cavity_2_name]_interaction.cross_kerr_coefficient`` (where
                ``[transmon_n_name]`` and  ``[cavity_n_name]`` are the names assigned to the
                respective transmons or cavities).
                The values are float/complex for constant coefficients and np.ndarray for
                piecewise-constant signals. If you pass a `cutoff_frequency`, the filtered
                versions of the piecewise-constant coefficients are also included with keys
                such as ``[transmon_2_name].drive_filtered``.
            ``infidelity``
                The state/operational infidelity of the optimized evolution.
            ``sample_times``
                The times at which the system's evolution is sampled,
                as an array of shape ``(T,)``.
            ``unitaries``
                The system's unitary time-evolution operators at each sample time,
                as an array of shape ``(T, D, D)``.
            ``state_evolution``
                The time evolution of the initial state at each sample time,
                as an array of shape ``(T, D)``.
                This is only returned if you provide an initial state.

    Notes
    -----
    The Hamiltonian of the system is of the form

    .. math::
        H = \\sum_i H_{\\mathrm{transmon}_i}
            + \\sum_i H_{\\mathrm{cavity}_i}
            + \\sum_{i,j} H_{\\mathrm{transmon}_i-\\mathrm{transmon}_j}
            + \\sum_{i,j} H_{\\mathrm{transmon}_i-\\mathrm{cavity}_j}
            + \\sum_{i,j} H_{\\mathrm{cavity}_i-\\mathrm{cavity}_j}

    where i and j mark the i-th and j-th transmon or cavity.
    For their definition of each Hamiltonian term, see its respective class.

    The Hilbert space of the system is defined as the outer product of all the
    transmon Hilbert spaces (in the order they're provided in `transmons`) with
    the cavity Hilbert spaces (in the order they're provided in `cavities`), that is:

    .. math::
        \\mathcal{H} =
            \\mathcal{H}_{\\mathrm{transmon}_1} \\otimes \\mathcal{H}_{\\mathrm{transmon}_2}
            \\otimes \\ldots
            \\otimes \\mathcal{H}_{\\mathrm{cavity}_1} \\otimes \\mathcal{H}_{\\mathrm{cavity}_2}
            \\otimes \\ldots

    The system dimension `D` is then the product of all transmon and cavity dimensions.

    If you provide an `initial_state` and a `target_state`, the optimization cost is defined as the
    infidelity of the state transfer process,

    .. math::
        \\mathcal{I}
            = 1 - \\left|
                \\langle
                    \\Psi_\\mathrm{target} | U(t_\\mathrm{gate}) | \\Psi_\\mathrm{initial}
                \\rangle
            \\right|^2 ,

    where :math:`U(t)` is the unitary time-evolution operator generated by the Hamiltonian.

    If you provide a `target_operation`, the optimization cost is defined as the operational
    infidelity,

    .. math::
        \\mathcal{I}
            = 1 - \\left| \\frac
                {\\mathrm{Tr} (U_\\mathrm{target}^\\dagger U(t_\\mathrm{gate}))}
                {\\mathrm{Tr} (U_\\mathrm{target}^\\dagger U_\\mathrm{target})}
            \\right|^2 .
    """

    check_argument(
        (target_state is None) ^ (target_operation is None),
        "You have to provide exactly one of `target_state` or `target_operation`.",
        {"target_state": target_state, "target_operation": target_operation},
    )

    system_dimension = np.prod(
        [system.dimension for system in transmons + cavities], dtype=int
    )

    if initial_state is not None:
        check_argument(
            initial_state.shape == (system_dimension,),
            "Initial state must be a 1D array of length "
            "np.prod([system.dimension for system in transmons + cavities]).",
            {
                "initial_state": initial_state,
                "transmons": transmons,
                "cavities": cavities,
            },
            extras={
                "initial_state.shape": initial_state.shape,
                "np.prod([system.dimension for system in transmons + cavities])": system_dimension,
            },
        )

    if target_state is not None:
        check_argument(
            initial_state is not None,
            "If you provide a `target_state`, you must provide an `initial_state`.",
            {"target_state": target_state, "initial_state": initial_state},
        )
        check_argument(
            target_state.shape == (system_dimension,),
            "Target state must be a 1D array of length "
            "np.prod([system.dimension for system in transmons + cavities]).",
            {
                "target_state": target_state,
                "transmons": transmons,
                "cavities": cavities,
            },
            extras={
                "target_state.shape": target_state.shape,
                "np.prod([system.dimension for system in transmons + cavities])": system_dimension,
            },
        )

    if target_operation is not None:
        check_argument(
            target_operation.shape == (system_dimension, system_dimension),
            "Target operation must be a square operator of shape (D, D) with "
            "D = np.prod([system.dimension for system in transmons + cavities]).",
            {
                "target_operation": target_operation,
                "transmons": transmons,
                "cavities": cavities,
            },
            extras={"target_operation.shape": target_operation.shape},
        )

    check_argument(
        gate_duration > 0,
        "The gate duration must be grater than zero.",
        {"gate_duration": gate_duration},
    )

    # Create graph object.
    graph = qctrl.create_graph()

    # Create PWC Hamiltonian.
    hamiltonian, optimizable_node_names = _create_superconducting_hamiltonian(
        graph=graph,
        transmons=transmons,
        cavities=cavities,
        interactions=interactions,
        gate_duration=gate_duration,
        cutoff_frequency=cutoff_frequency,
        sample_count=sample_count,
    )

    # Check whether there are any optimizable coefficients.
    check_argument(
        len(optimizable_node_names) > 0,
        "At least one of the Hamiltonian terms must be optimizable.",
        {"transmons": transmons, "cavities": cavities, "interactions": interactions},
    )

    other_output_node_names = ["unitaries", "infidelity"]

    # Calculate the evolution.
    sample_times = np.linspace(
        gate_duration / sample_count, gate_duration, sample_count
    )
    unitaries = graph.time_evolution_operators_pwc(
        hamiltonian=hamiltonian, sample_times=sample_times, name="unitaries"
    )

    if initial_state is not None:
        states = unitaries @ initial_state[:, None]
        states = states[..., 0]
        states.name = "state_evolution"
        other_output_node_names.append("state_evolution")

    # Calculate the infidelity.
    if target_state is not None:
        graph.state_infidelity(target_state, states[-1], name="infidelity")
    else:
        graph.unitary_infidelity(unitaries[-1], target_operation, name="infidelity")

    # Execute optimization.
    optimization_result = qctrl.functions.calculate_optimization(
        graph=graph,
        cost_node_name="infidelity",
        output_node_names=optimizable_node_names + other_output_node_names,
        target_cost=target_cost,
        optimization_count=optimization_count,
    )

    # Retrieve results and build output dictionary.
    result_dict = {"optimized_values": {}, "sample_times": sample_times}

    for key in optimizable_node_names:
        result_dict["optimized_values"][key] = _extract_output(
            optimization_result.output[key]
        )

    for key in other_output_node_names:
        result_dict[key] = _extract_output(optimization_result.output[key])

    return result_dict
